from typing import List
from typing import Tuple

from ewokscore import Task
from ewokscore.model import BaseOutputModel
from pydantic import Field

from ..models import FluoInputs
from .scan_data import save_2d_xrf_scans

SEED = 100
MAX_DEVIATION = 0.05  # Fraction of a single step


class MeshSingleScanSingleDetectorInputs(FluoInputs):
    output_filename: str = Field(
        description="Bliss dataset HDF5 file name.", examples=["/data/dataset.h5"]
    )
    emission_line_groups: List[str] = Field(
        default_factory=lambda: ["Si-K", "Ca-K", "Ce-L", "Fe-K"]
    )
    energy: float = Field(default=12.0)
    shape: Tuple[int, int] = Field(default=(50, 60))
    expo_time: float = Field(default=0.1)
    flux: float = Field(default=1e7)
    counting_noise: bool = Field(default=True)
    rois: List[Tuple[int, int]] = Field(
        default_factory=lambda: [(100, 200), (300, 600)]
    )
    integral_type: bool = Field(default=True)


class MeshSingleScanSingleDetectorOutputs(BaseOutputModel):
    filename: str = Field(
        description="Bliss dataset HDF5 file name.", examples=["/data/dataset.h5"]
    )
    scan_number: int = Field(description="Scan number.", examples=[1, 2])
    config: str
    expo_time: float
    monitor_name: str
    monitor_normalization_template: str
    detector_name: str
    detector_normalization_template: str


class MeshSingleScanSingleDetector(
    Task,
    input_model=MeshSingleScanSingleDetectorInputs,
    output_model=MeshSingleScanSingleDetectorOutputs,
):
    """XRF test data of one scan with one detector"""

    def run(self):
        scan_number = 1
        filename = self.inputs.output_filename

        _ = save_2d_xrf_scans(
            filename=filename,
            emission_line_groups=self.inputs.emission_line_groups,
            first_scan_number=scan_number,
            shape=self.inputs.shape,
            mosaic=(1, 1),
            energy=self.inputs.energy,
            flux=self.inputs.flux,
            expo_time=self.inputs.expo_time,
            counting_noise=self.inputs.counting_noise,
            integral_type=self.inputs.integral_type,
            rois=self.inputs.rois,
            nmcas=1,
            max_deviation=MAX_DEVIATION,
            seed=SEED,
        )

        self.outputs.filename = filename
        self.outputs.scan_number = scan_number
        self.outputs.config = f"{filename}::/{scan_number}.1/theory/configuration/data"
        self.outputs.expo_time = self.inputs.expo_time
        self.outputs.monitor_name = "I0"
        self.outputs.monitor_normalization_template = (
            f"{int(self.inputs.flux * self.inputs.expo_time)}/<instrument/{{}}/data>"
        )
        self.outputs.detector_name = "mca0"
        self.outputs.detector_normalization_template = (
            f"{self.inputs.expo_time}/<instrument/{{}}/live_time>"
        )


class MeshSingleScanMultiDetectorInputs(FluoInputs):
    output_filename: str = Field(
        description="Bliss dataset HDF5 file name.", examples=["/data/dataset.h5"]
    )
    emission_line_groups: List[str] = Field(
        default_factory=lambda: ["Si-K", "Ca-K", "Ce-L", "Fe-K"]
    )
    energy: float = Field(default=12.0)
    shape: Tuple[int, int] = Field(default=(50, 60))
    expo_time: float = Field(default=0.1)
    ndetectors: int = Field(default=2)
    flux: float = Field(default=1e7)
    counting_noise: bool = Field(default=True)
    rois: List[Tuple[int, int]] = Field(
        default_factory=lambda: [(100, 200), (300, 600)]
    )
    integral_type: bool = Field(default=True)


class MeshSingleScanMultiDetectorOutputs(BaseOutputModel):
    filename: str = Field(
        description="Bliss dataset HDF5 file name.", examples=["/data/dataset.h5"]
    )
    scan_number: int = Field(description="Scan number.", examples=[1, 2])
    config: str
    configs: List[str]
    expo_time: float
    monitor_name: str
    monitor_normalization_template: str
    detector_names: List[str]
    detector_normalization_template: str


class MeshSingleScanMultiDetector(
    Task,
    input_model=MeshSingleScanMultiDetectorInputs,
    output_model=MeshSingleScanMultiDetectorOutputs,
):
    """XRF test data of one scan with multiple detectors"""

    def run(self):
        scan_number = 1
        filename = self.inputs.output_filename

        _ = save_2d_xrf_scans(
            filename=filename,
            emission_line_groups=self.inputs.emission_line_groups,
            first_scan_number=scan_number,
            shape=self.inputs.shape,
            mosaic=(1, 1),
            energy=self.inputs.energy,
            flux=self.inputs.flux,
            expo_time=self.inputs.expo_time,
            counting_noise=self.inputs.counting_noise,
            integral_type=self.inputs.integral_type,
            rois=self.inputs.rois,
            nmcas=self.inputs.ndetectors,
            max_deviation=MAX_DEVIATION,
            seed=SEED,
        )

        self.outputs.filename = filename
        self.outputs.scan_number = scan_number
        self.outputs.configs = [
            f"{filename}::/{scan_number}.1/theory/configuration/data"
        ] * self.inputs.ndetectors
        self.outputs.config = f"{filename}::/{scan_number}.1/theory/configuration/data"
        self.outputs.expo_time = self.inputs.expo_time
        self.outputs.monitor_name = "I0"
        self.outputs.monitor_normalization_template = (
            f"{int(self.inputs.flux * self.inputs.expo_time)}/<instrument/{{}}/data>"
        )
        self.outputs.detector_names = [f"mca{i}" for i in range(self.inputs.ndetectors)]
        self.outputs.detector_normalization_template = (
            f"{self.inputs.expo_time}/<instrument/{{}}/live_time>"
        )


class MosaicMeshSingleDetectorInputs(FluoInputs):
    output_filename: str = Field(
        description="Bliss dataset HDF5 file name.", examples=["/data/dataset.h5"]
    )
    emission_line_groups: List[str] = Field(
        default_factory=lambda: ["Si-K", "Ca-K", "Ce-L", "Fe-K"]
    )
    energy: float = Field(default=12.0)
    shape: Tuple[int, int] = Field(default=(50, 60))
    mosaic: Tuple[int, int] = Field(default=(2, 3))
    expo_time: float = Field(default=0.1)
    ndetectors: int = Field(default=1)
    flux: float = Field(default=1e7)
    counting_noise: bool = Field(default=True)
    rois: List[Tuple[int, int]] = Field(
        default_factory=lambda: [(100, 200), (300, 600)]
    )
    integral_type: bool = Field(default=True)


class MosaicMeshSingleDetectorOutputs(BaseOutputModel):
    filenames: List[str]
    scan_ranges: List[Tuple[int, int]]
    config: str
    expo_time: float
    monitor_name: str
    monitor_normalization_template: str
    detector_name: str
    detector_normalization_template: str


class MosaicMeshSingleDetector(
    Task,
    input_model=MosaicMeshSingleDetectorInputs,
    output_model=MosaicMeshSingleDetectorOutputs,
):
    """XRF test data of a mosaic scan with one detector"""

    def run(self):
        filename = self.inputs.output_filename

        scan_numbers = save_2d_xrf_scans(
            filename=filename,
            emission_line_groups=self.inputs.emission_line_groups,
            first_scan_number=1,
            shape=self.inputs.shape,
            mosaic=self.inputs.mosaic,
            energy=self.inputs.energy,
            flux=self.inputs.flux,
            expo_time=self.inputs.expo_time,
            counting_noise=self.inputs.counting_noise,
            integral_type=self.inputs.integral_type,
            rois=self.inputs.rois,
            nmcas=1,
            max_deviation=MAX_DEVIATION,
            seed=SEED,
        )

        self.outputs.filenames = [filename]
        self.outputs.scan_ranges = [[scan_numbers[0], scan_numbers[-1]]]
        self.outputs.config = (
            f"{filename}::/{scan_numbers[0]}.1/theory/configuration/data"
        )
        self.outputs.expo_time = self.inputs.expo_time
        self.outputs.monitor_name = "I0"
        self.outputs.monitor_normalization_template = (
            f"{int(self.inputs.flux * self.inputs.expo_time)}/<instrument/{{}}/data>"
        )
        self.outputs.detector_name = "mca0"
        self.outputs.detector_normalization_template = (
            f"{self.inputs.expo_time}/<instrument/{{}}/live_time>"
        )


class MosaicMeshMultiDetectorInputs(FluoInputs):
    output_filename: str = Field(
        description="Bliss dataset HDF5 file name.", examples=["/data/dataset.h5"]
    )
    emission_line_groups: List[str] = Field(
        default_factory=lambda: ["Si-K", "Ca-K", "Ce-L", "Fe-K"]
    )
    energy: float = Field(default=12.0)
    shape: Tuple[int, int] = Field(default=(50, 60))
    mosaic: Tuple[int, int] = Field(default=(2, 3))
    expo_time: float = Field(default=0.1)
    ndetectors: int = Field(default=2)
    flux: float = Field(default=1e7)
    counting_noise: bool = Field(default=True)
    rois: List[Tuple[int, int]] = Field(
        default_factory=lambda: [(100, 200), (300, 600)]
    )
    integral_type: bool = Field(default=True)


class MosaicMeshMultiDetectorOutputs(BaseOutputModel):
    filenames: List[str]
    scan_ranges: List[Tuple[int, int]]
    configs: List[str]
    config: str
    expo_time: float
    monitor_name: str
    monitor_normalization_template: str
    detector_names: List[str]
    detector_normalization_template: str


class MosaicMeshMultiDetector(
    Task,
    input_model=MosaicMeshMultiDetectorInputs,
    output_model=MosaicMeshMultiDetectorOutputs,
):
    """XRF test data of a mosaic scan with multiple detectors"""

    def run(self):
        filename = self.inputs.output_filename

        scan_numbers = save_2d_xrf_scans(
            filename=filename,
            emission_line_groups=self.inputs.emission_line_groups,
            first_scan_number=1,
            shape=self.inputs.shape,
            mosaic=self.inputs.mosaic,
            energy=self.inputs.energy,
            flux=self.inputs.flux,
            expo_time=self.inputs.expo_time,
            counting_noise=self.inputs.counting_noise,
            integral_type=self.inputs.integral_type,
            rois=self.inputs.rois,
            nmcas=self.inputs.ndetectors,
            max_deviation=MAX_DEVIATION,
            seed=SEED,
        )

        self.outputs.filenames = [filename]
        self.outputs.scan_ranges = [[scan_numbers[0], scan_numbers[-1]]]
        self.outputs.configs = [
            f"{filename}::/{scan_numbers[0]}.1/theory/configuration/data"
        ] * self.inputs.ndetectors
        self.outputs.config = (
            f"{filename}::/{scan_numbers[0]}.1/theory/configuration/data"
        )
        self.outputs.expo_time = self.inputs.expo_time
        self.outputs.monitor_name = "I0"
        self.outputs.monitor_normalization_template = (
            f"{int(self.inputs.flux * self.inputs.expo_time)}/<instrument/{{}}/data>"
        )
        self.outputs.detector_names = [f"mca{i}" for i in range(self.inputs.ndetectors)]
        self.outputs.detector_normalization_template = (
            f"{self.inputs.expo_time}/<instrument/{{}}/live_time>"
        )


class MeshStackSingleDetectorInputs(FluoInputs):
    output_filename: str = Field(
        description="Bliss dataset HDF5 file name.", examples=["/data/dataset.h5"]
    )
    nscans: int = Field(default=3)
    emission_line_groups: List[str] = Field(
        default_factory=lambda: ["Si-K", "Ca-K", "Ce-L", "Fe-K"]
    )
    energy: float = Field(default=12.0)
    shape: Tuple[int, int] = Field(default=(50, 60))
    expo_time: float = Field(default=0.1)
    flux: float = Field(default=1e7)
    counting_noise: bool = Field(default=True)
    rois: List[Tuple[int, int]] = Field(
        default_factory=lambda: [(100, 200), (300, 600)]
    )
    integral_type: bool = Field(default=True)


class MeshStackSingleDetectorOutputs(BaseOutputModel):
    filenames: List[str] = Field(
        description="Bliss dataset HDF5 file name.",
        examples=[["/data/dataset1.h5", "/data/dataset2.h5"]],
    )
    scan_ranges: List[Tuple[int, int]]
    config: str
    expo_time: float
    monitor_name: str
    monitor_normalization_template: str
    detector_name: str
    detector_normalization_template: str


class MeshStackSingleDetector(
    Task,
    input_model=MeshStackSingleDetectorInputs,
    output_model=MeshStackSingleDetectorOutputs,
):
    """XRF test data of a stack of identical scans with one detector"""

    def run(self):
        filename = self.inputs.output_filename

        bliss_scan_uris = list()
        scan_numbers = list(range(1, self.inputs.nscans + 1))
        energy = self.inputs.energy
        for scan_number in scan_numbers:
            _ = save_2d_xrf_scans(
                filename=filename,
                emission_line_groups=self.inputs.emission_line_groups,
                first_scan_number=scan_number,
                shape=self.inputs.shape,
                mosaic=(1, 1),
                energy=energy,
                flux=self.inputs.flux,
                expo_time=self.inputs.expo_time,
                counting_noise=self.inputs.counting_noise,
                integral_type=self.inputs.integral_type,
                rois=self.inputs.rois,
                nmcas=1,
                max_deviation=MAX_DEVIATION,
                seed=SEED,
            )

            bliss_scan_uris.append(f"{filename}::/{scan_number}.1")
            energy += 0.010

        self.outputs.filenames = [filename]
        self.outputs.scan_ranges = [[scan_numbers[0], scan_numbers[-1]]]
        self.outputs.config = (
            f"{filename}::/{scan_numbers[0]}.1/theory/configuration/data"
        )
        self.outputs.monitor_name = "I0"
        self.outputs.monitor_normalization_template = (
            f"{int(self.inputs.flux * self.inputs.expo_time)}/<instrument/{{}}/data>"
        )
        self.outputs.expo_time = self.inputs.expo_time
        self.outputs.detector_name = "mca0"
        self.outputs.detector_normalization_template = (
            f"{self.inputs.expo_time}/<instrument/{{}}/live_time>"
        )


class MeshStackMultiDetectorInputs(FluoInputs):
    output_filename: str = Field(
        description="Bliss dataset HDF5 file name.", examples=["/data/dataset.h5"]
    )
    nscans: int = Field(default=3)
    emission_line_groups: List[str] = Field(
        default_factory=lambda: ["Si-K", "Ca-K", "Ce-L", "Fe-K"]
    )
    energy: float = Field(default=12.0)
    shape: Tuple[int, int] = Field(default=(50, 60))
    expo_time: float = Field(default=0.1)
    ndetectors: int = Field(default=2)
    flux: float = Field(default=1e7)
    counting_noise: bool = Field(default=True)
    rois: List[Tuple[int, int]] = Field(
        default_factory=lambda: [(100, 200), (300, 600)]
    )
    integral_type: bool = Field(default=True)


class MeshStackMultiDetectorOutputs(BaseOutputModel):
    filenames: List[str] = Field(
        description="Bliss dataset HDF5 file name.",
        examples=[["/data/dataset1.h5", "/data/dataset2.h5"]],
    )
    scan_ranges: List[Tuple[int, int]]
    configs: List[str]
    config: str
    expo_time: float
    monitor_name: str
    monitor_normalization_template: str
    detector_names: List[str]
    detector_normalization_template: str


class MeshStackMultiDetector(
    Task,
    input_model=MeshStackMultiDetectorInputs,
    output_model=MeshStackMultiDetectorOutputs,
):
    """XRF test data of a stack of identical scans with multiple detectors"""

    def run(self):
        filename = self.inputs.output_filename

        bliss_scan_uris = list()
        scan_numbers = list(range(1, self.inputs.nscans + 1))
        energy = self.inputs.energy
        for scan_number in scan_numbers:
            _ = save_2d_xrf_scans(
                filename=filename,
                emission_line_groups=self.inputs.emission_line_groups,
                first_scan_number=scan_number,
                shape=self.inputs.shape,
                mosaic=(1, 1),
                energy=energy,
                flux=self.inputs.flux,
                expo_time=self.inputs.expo_time,
                counting_noise=self.inputs.counting_noise,
                integral_type=self.inputs.integral_type,
                rois=self.inputs.rois,
                nmcas=self.inputs.ndetectors,
                max_deviation=MAX_DEVIATION,
                seed=SEED,
            )

            bliss_scan_uris.append(f"{filename}::/{scan_number}.1")
            energy += 0.010

        self.outputs.filenames = [filename]
        self.outputs.scan_ranges = [[scan_numbers[0], scan_numbers[-1]]]
        self.outputs.configs = [
            f"{filename}::/{scan_numbers[0]}.1/theory/configuration/data"
        ] * self.inputs.ndetectors
        self.outputs.config = (
            f"{filename}::/{scan_numbers[0]}.1/theory/configuration/data"
        )
        self.outputs.monitor_name = "I0"
        self.outputs.monitor_normalization_template = (
            f"{int(self.inputs.flux * self.inputs.expo_time)}/<instrument/{{}}/data>"
        )
        self.outputs.expo_time = self.inputs.expo_time
        self.outputs.detector_names = [f"mca{i}" for i in range(self.inputs.ndetectors)]
        self.outputs.detector_normalization_template = (
            f"{self.inputs.expo_time}/<instrument/{{}}/live_time>"
        )



import logging
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import numpy
from scipy.optimize import minimize

from . import grid_utils
from . import rounding

logger = logging.getLogger(__name__)


def optimal_grid_axes(
    scatter_coordinates: Sequence[numpy.ndarray],
    fix_resolution: bool = True,
    fix_limits: bool = True,
    resolution: Optional[Sequence[float]] = None,
) -> List[numpy.ndarray]:
    """
    :param scatter_coordinates: shape `(Nscatter, Ndim)`
    :param fixed_resolution:
    :param fixed_limits:
    :param resolution:
    :returns: `Ndim` 1D arrays with length `N0`, `N1`, ...
    """
    # Validate parameters
    if scatter_coordinates.ndim != 2:
        raise ValueError(
            "Scatter coordinates is not a 2D array with shape (Nscatter, Ndim)"
        )
    num_data_points, Ndim = scatter_coordinates.shape
    if resolution is None:
        resolution = [resolution] * Ndim

    # Check for dimensions with zero data range
    min_limits = numpy.min(scatter_coordinates, axis=0)
    max_limits = numpy.max(scatter_coordinates, axis=0)
    max_data_range = max_limits - min_limits
    zero_dimensions = numpy.where(max_data_range == 0)[0]

    if zero_dimensions.any():
        zero_dimensions = zero_dimensions.tolist()
        rmin = min_limits[zero_dimensions]
        rmax = max_limits[zero_dimensions]
        range_strings = [
            f"Dim {i}: {r[0]} → {r[1]}" for i, r in enumerate(zip(rmin, rmax))
        ]
        range_report = "\n ".join(range_strings)
        raise ValueError(
            f"Cannot compute grid axes: the following dimension(s) have zero data range\n {range_report}"
        )

    # Find optimal grid analytically based and estimate
    # the optimal resolution when not provided.
    start_stop_size = (
        _estimate_start_stop_size(scatter_coordinates[:, i], resolution=resolutioni)
        for i, resolutioni in enumerate(resolution)
    )
    initial_start_limits, initial_stop_limits, initial_grid_shape = zip(
        *start_stop_size
    )

    logger.debug("Initial grid shape: %s", initial_grid_shape)

    initial_start_limits = numpy.array(initial_start_limits)
    initial_stop_limits = numpy.array(initial_stop_limits)
    initial_data_range = initial_stop_limits - initial_start_limits

    # Compile parameters of optimization
    initial_guess = []
    bounds = []
    fixed_limits = []
    fixed_resolution = []

    initial_resolution = initial_data_range / (numpy.array(initial_grid_shape) - 1)
    logger.info("Initial resolution: %s", tuple(map(float, initial_resolution)))

    if fix_resolution:
        fixed_resolution.extend(initial_resolution)
    else:
        initial_guess.extend(initial_resolution)
        resolution_bounds = zip(max_data_range / (num_data_points - 1), max_data_range)
        bounds.extend(resolution_bounds)

    if fix_limits:
        fixed_limits.extend(initial_start_limits)
        fixed_limits.extend(initial_stop_limits)
    else:
        initial_guess.extend(initial_start_limits)
        initial_guess.extend(initial_stop_limits)
        start_limit_bounts = zip(min_limits, max_limits)
        bounds.extend(start_limit_bounts)
        stop_limit_bounts = zip(max_limits, max_limits)
        bounds.extend(stop_limit_bounts)

    if fixed_limits:
        fixed_limits = numpy.asarray(fixed_limits)
    else:
        fixed_limits = None
    if fixed_resolution:
        fixed_resolution = numpy.asarray(fixed_resolution)
    else:
        fixed_resolution = None

    # Do optimization unless all parameters are fixed
    if initial_guess:
        logger.info("Grid optimization applied")
        initial_guess = numpy.asarray(initial_guess)
        result = minimize(
            _scatter_grid_distance_measure,
            initial_guess,
            args=(scatter_coordinates, fixed_resolution, fixed_limits),
            method="Nelder-Mead",
            bounds=bounds,
        )
        fit_parameters = result.x
    else:
        logger.info("No grid optimization applied")
        fit_parameters = None

    grid_axes = _fit_parameters_to_grid_axes(
        Ndim,
        fit_parameters=fit_parameters,
        fixed_resolution=fixed_resolution,
        fixed_limits=fixed_limits,
    )

    final_grid_shape = tuple(len(arr) for arr in grid_axes)
    final_resolution = tuple(
        arr[1] - arr[0] if arr.size else numpy.nan for arr in grid_axes
    )
    logger.info("Final grid shape: %s", final_grid_shape)
    logger.info("Final resolution: %s", tuple(map(float, final_resolution)))
    return grid_axes


def _estimate_start_stop_size(
    array: numpy.ndarray, resolution: Optional[float] = None
) -> Tuple[float, float, int]:
    """
    :params array: 1D array
    :param resolution: bin width
    :returns: start value, stop value, number of points
    """
    if len(array) == 0:
        return numpy.nan, numpy.nan, 0
    if len(array) == 1:
        return array[0], array[0], 1

    array = numpy.sort(array)

    if not resolution:
        distances = numpy.diff(array)

        min_distance = numpy.min(distances)
        mid_distance = min_distance + (numpy.max(distances) - min_distance) / 2

        mask = distances > mid_distance
        if mask.any():
            resolution = numpy.median(distances[mask])
        else:
            resolution = numpy.median(distances)
        resolution = rounding.round_to_significant(resolution)

    start, stop, nbins = rounding.round_range(array[0], array[-1], resolution)
    if nbins == 0:
        return start, stop, 0
    npoints = nbins + 1
    return start, stop, npoints


def _scatter_grid_distance_measure(
    fit_parameters: numpy.ndarray,
    scatter_coordinates: numpy.ndarray,
    fixed_resolution: Optional[numpy.ndarray] = None,
    fixed_limits: Optional[numpy.ndarray] = None,
) -> float:
    """Distance measure between scattered points and a grid with a resolution and
    start and stop limits that are either refined or fixed.

    :param fit_parameters:
    :param scatter_coordinates: `(Nscatter, Ndim)`
    :param fixed_resolution:
    :param fixed_limits:
    :returns: sum of closest distances of grid nodes with scatter_coordinates
    """
    _, Ndim = scatter_coordinates.shape

    grid_axes = _fit_parameters_to_grid_axes(
        Ndim,
        fit_parameters=fit_parameters,
        fixed_limits=fixed_limits,
        fixed_resolution=fixed_resolution,
    )  # [(N0,), (N1,), ...]

    expanded_grid_coordinates = grid_utils.expanded_grid_coordinates(
        grid_axes
    )  # (N0*N1*... , Ndim)

    # Closest grid point for each scatter point
    _, scatter_distances = grid_utils.closest_point(
        expanded_grid_coordinates, scatter_coordinates
    )
    scatter_coverage = numpy.sum(scatter_distances)

    # Closest scatter point for each grid point
    _, grid_distances = grid_utils.closest_point(
        scatter_coordinates, expanded_grid_coordinates
    )
    grid_coverage = numpy.sum(grid_distances)

    return scatter_coverage + grid_coverage


def _fit_parameters_to_grid_axes(
    Ndim: int,
    fit_parameters: Optional[numpy.ndarray] = None,
    fixed_resolution: Optional[numpy.ndarray] = None,
    fixed_limits: Optional[numpy.ndarray] = None,
) -> List[numpy.ndarray]:
    """
    :param fit_parameters:
    :param fixed_resolution:
    :param fixed_limits:
    :returns: `Ndim` arrays with shape `(N0,)`, `(N1,)`, ...
    """
    fix_resolution = fixed_resolution is not None
    fix_limits = fixed_limits is not None

    params_offset = 0

    if fix_resolution:
        grid_resolution = fixed_resolution
    else:
        grid_resolution = fit_parameters[params_offset : params_offset + Ndim]
        params_offset += Ndim

    if fix_limits:
        grid_start_limits = fixed_limits[:Ndim]
        grid_stop_limits = fixed_limits[Ndim:]
    else:
        grid_start_limits = fit_parameters[params_offset : params_offset + Ndim]
        params_offset += Ndim
        grid_stop_limits = fit_parameters[params_offset : params_offset + Ndim]
        params_offset += Ndim

    _linspace = _linspace_fixed_stop

    grid_axes = list()
    for i, (start, stop, step) in enumerate(
        zip(grid_start_limits, grid_stop_limits, grid_resolution)
    ):
        if step == 0 or not numpy.isfinite(start) or not numpy.isfinite(stop):
            raise ValueError(
                f"The grid along dimension {i} cannot be determined ({start} → {stop})"
            )
        grid_axes.append(_linspace(start, stop, step))
    return grid_axes


def _linspace_fixed_step(start: float, stop: float, step: float) -> numpy.ndarray:
    return numpy.arange(start, stop, step)


def _linspace_fixed_stop(start: float, stop: float, step: float) -> numpy.ndarray:
    n = max(int((stop - start) / step + 1.5), 1)
    return numpy.linspace(start, stop, n)



import h5py
import pytest
from ewoksorange.tests.utils import execute_task

from orangecontrib.ewoksfluo.categories.demo.mesh_single_scan_multi_det import (
    OWMeshSingleScanMultiDetector,
)
from orangecontrib.ewoksfluo.categories.demo.mesh_single_scan_single_det import (
    OWMeshSingleScanSingleDetector,
)
from orangecontrib.ewoksfluo.categories.demo.mesh_stack_multi_det import (
    OWMeshStackMultiDetector,
)
from orangecontrib.ewoksfluo.categories.demo.mesh_stack_single_det import (
    OWMeshStackSingleDetector,
)
from orangecontrib.ewoksfluo.categories.demo.mosaic_mesh_multi_det import (
    OWMosaicMeshMultiDetector,
)
from orangecontrib.ewoksfluo.categories.demo.mosaic_mesh_single_det import (
    OWMosaicMeshSingleDetector,
)

from .. import utils

_TASK_CLASSES = {
    OWMeshSingleScanSingleDetector: {"nscans": 1, "ndetectors": 1, "mosaic": None},
    OWMeshSingleScanMultiDetector: {"nscans": 1, "ndetectors": 2, "mosaic": None},
    OWMeshStackSingleDetector: {"nscans": 2, "ndetectors": 1, "mosaic": None},
    OWMeshStackMultiDetector: {"nscans": 2, "ndetectors": 2, "mosaic": None},
    OWMosaicMeshSingleDetector: {"nscans": 1, "ndetectors": 1, "mosaic": (2, 3)},
    OWMosaicMeshMultiDetector: {"nscans": 1, "ndetectors": 2, "mosaic": (2, 3)},
}


@pytest.mark.parametrize("task_cls", _TASK_CLASSES)
def test_mesh_tasks(tmp_path, task_cls):
    _test_example_xrf_scan(tmp_path, task_cls)


@pytest.mark.parametrize("task_cls", _TASK_CLASSES)
def test_mesh_tasks_widget(tmp_path, qtapp, task_cls):
    _test_example_xrf_scan(tmp_path, task_cls, widget=True)


def _test_example_xrf_scan(tmp_path, task_cls, widget: bool = False):
    params = _TASK_CLASSES[task_cls]
    nscans = params["nscans"]
    ndetectors = params["ndetectors"]
    mosaic = params["mosaic"]

    if not widget:
        task_cls = task_cls.ewokstaskclass
    filename = str(tmp_path / "test.h5")
    inputs = {
        "output_filename": filename,
        "rois": [(100, 200), (300, 600)],
        "emission_line_groups": ["Si-K", "Ca-K", "Ce-L", "Fe-K"],
    }
    if ndetectors > 1:
        inputs["ndetectors"] = ndetectors
    if nscans > 1:
        inputs["nscans"] = nscans
    if mosaic:
        inputs["mosaic"] = mosaic
        npoints = 500  # 50//mosaic[0] * 60//mosaic[1]
        nscans_total = nscans * mosaic[0] * mosaic[1]
    else:
        npoints = 3000  # 50 * 60
        nscans_total = nscans

    for _ in range(2):  # Repeat twice to test overwrite
        outputs = execute_task(task_cls, inputs)
        assert outputs == expected_task_outputs(nscans, ndetectors, mosaic, tmp_path)

        with h5py.File(filename) as f:
            content = utils.h5content(f)

        expected_content = expected_h5content(nscans_total, npoints, ndetectors)

        assert content == expected_content


def expected_task_outputs(nscans, ndetectors, mosaic, tmp_path):
    filename = str(tmp_path / "test.h5")
    outputs = {
        "config": f"{filename}::/1.1/theory/configuration/data",
        "expo_time": 0.1,
        "monitor_name": "I0",
    }
    if mosaic and ndetectors == 1:
        nscans = mosaic[0] * mosaic[1]
        return {
            "detector_name": "mca0",
            "filenames": [filename],
            "scan_ranges": [(1, nscans)],
            "monitor_normalization_template": "1000000/<instrument/{}/data>",
            "detector_normalization_template": "0.1/<instrument/{}/live_time>",
            **outputs,
        }
    if mosaic and ndetectors > 1:
        nscans = mosaic[0] * mosaic[1]
        return {
            "configs": [f"{filename}::/1.1/theory/configuration/data"] * ndetectors,
            "detector_names": ["mca0", "mca1"],
            "filenames": [filename],
            "scan_ranges": [(1, nscans)],
            "monitor_normalization_template": "1000000/<instrument/{}/data>",
            "detector_normalization_template": "0.1/<instrument/{}/live_time>",
            **outputs,
        }
    if nscans == 1 and ndetectors == 1:
        return {
            "detector_name": "mca0",
            "filename": filename,
            "scan_number": 1,
            "monitor_normalization_template": "1000000/<instrument/{}/data>",
            "detector_normalization_template": "0.1/<instrument/{}/live_time>",
            **outputs,
        }
    if nscans > 1 and ndetectors == 1:
        return {
            "detector_name": "mca0",
            "filenames": [filename],
            "scan_ranges": [(1, nscans)],
            "monitor_normalization_template": "1000000/<instrument/{}/data>",
            "detector_normalization_template": "0.1/<instrument/{}/live_time>",
            **outputs,
        }
    if nscans == 1 and ndetectors > 1:
        return {
            "configs": [f"{filename}::/1.1/theory/configuration/data"] * ndetectors,
            "detector_names": ["mca0", "mca1"],
            "filename": filename,
            "scan_number": 1,
            "monitor_normalization_template": "1000000/<instrument/{}/data>",
            "detector_normalization_template": "0.1/<instrument/{}/live_time>",
            **outputs,
        }
    if nscans > 1 and ndetectors > 1:
        return {
            "configs": [f"{filename}::/1.1/theory/configuration/data"] * ndetectors,
            "detector_names": ["mca0", "mca1"],
            "filenames": [filename],
            "scan_ranges": [(1, nscans)],
            "monitor_normalization_template": "1000000/<instrument/{}/data>",
            "detector_normalization_template": "0.1/<instrument/{}/live_time>",
            **outputs,
        }


def expected_h5content(nscans, npoints, ndetectors):
    content = {"@attrs": {"NX_class", "creator"}}

    mca_detector = {
        "@attrs": {"NX_class"},
        "data@shape": (npoints, 1024),
        "elapsed_time@shape": (npoints,),
        "event_count_rate@shape": (npoints,),
        "events@shape": (npoints,),
        "fractional_dead_time@shape": (npoints,),
        "live_time@shape": (npoints,),
        "roi1@shape": (npoints,),
        "roi2@shape": (npoints,),
        "spectrum@shape": (npoints, 1024),
        "trigger_count_rate@shape": (npoints,),
        "trigger_live_time@shape": (npoints,),
        "triggers@shape": (npoints,),
    }

    mca_meas = {
        "mca{detector}@shape": (npoints, 1024),
        "mca{detector}_elapsed_time@shape": (npoints,),
        "mca{detector}_event_count_rate@shape": (npoints,),
        "mca{detector}_events@shape": (npoints,),
        "mca{detector}_fractional_dead_time@shape": (npoints,),
        "mca{detector}_live_time@shape": (npoints,),
        "mca{detector}_roi1@shape": (npoints,),
        "mca{detector}_roi2@shape": (npoints,),
        "mca{detector}_trigger_count_rate@shape": (npoints,),
        "mca{detector}_trigger_live_time@shape": (npoints,),
        "mca{detector}_triggers@shape": (npoints,),
    }

    for scan in range(1, nscans + 1):
        scan_content = {
            "@attrs": {"NX_class"},
            "instrument": {
                "@attrs": {"NX_class"},
                "I0": {"@attrs": {"NX_class"}, "data@shape": (npoints,)},
                "positioners": {
                    "@attrs": {"NX_class"},
                    "energy@attrs": {"units"},
                    "energy@shape": (),
                    "sampy@attrs": {"units"},
                    "sampy@shape": (npoints,),
                    "sampz@attrs": {"units"},
                    "sampz@shape": (npoints,),
                },
                "positioners_start": {
                    "@attrs": {"NX_class"},
                    "energy@attrs": {"units"},
                    "energy@shape": (),
                    "sampy@attrs": {"units"},
                    "sampy@shape": (),
                    "sampz@attrs": {"units"},
                    "sampz@shape": (),
                },
                "positioners_end": {
                    "@attrs": {"NX_class"},
                    "energy@attrs": {"units"},
                    "energy@shape": (),
                    "sampy@attrs": {"units"},
                    "sampy@shape": (),
                    "sampz@attrs": {"units"},
                    "sampz@shape": (),
                },
                "sampy": {
                    "@attrs": {"NX_class"},
                    "value@attrs": {"units"},
                    "value@shape": (npoints,),
                },
                "sampz": {
                    "@attrs": {"NX_class"},
                    "value@attrs": {"units"},
                    "value@shape": (npoints,),
                },
            },
            "measurement": {
                "@attrs": {"NX_class"},
                "I0@shape": (npoints,),
                "sampy@attrs": {"units"},
                "sampy@shape": (npoints,),
                "sampz@attrs": {"units"},
                "sampz@shape": (npoints,),
            },
            "theory": {
                "@attrs": {"NX_class"},
                "configuration": {
                    "@attrs": {"NX_class"},
                    "data@shape": (),
                    "type@shape": (),
                },
                "description": {
                    "@attrs": {"NX_class"},
                    "data@shape": (),
                    "type@shape": (),
                },
                "I0_reference@shape": (),
                "parameters": {
                    "@attrs": {"NX_class", "auxiliary_signals", "axes", "signal"},
                    "Ca-K@shape": (npoints,),
                    "Ce-L@shape": (npoints,),
                    "Compton000@shape": (npoints,),
                    "Fe-K@shape": (npoints,),
                    "Peak000@shape": (npoints,),
                    "Si-K@shape": (npoints,),
                    "sampy@attrs": {"units"},
                    "sampy@shape": (npoints,),
                    "sampz@attrs": {"units"},
                    "sampz@shape": (npoints,),
                },
                "parameters_norm": {
                    "@attrs": {"NX_class", "auxiliary_signals", "axes", "signal"},
                    "Ca-K@shape": (npoints,),
                    "Ce-L@shape": (npoints,),
                    "Compton000@shape": (npoints,),
                    "Fe-K@shape": (npoints,),
                    "Peak000@shape": (npoints,),
                    "Si-K@shape": (npoints,),
                    "sampy@attrs": {"units"},
                    "sampy@shape": (npoints,),
                    "sampz@attrs": {"units"},
                    "sampz@shape": (npoints,),
                },
                "rois": {
                    "@attrs": {"NX_class", "auxiliary_signals", "axes", "signal"},
                    "roi1@shape": (npoints,),
                    "roi2@shape": (npoints,),
                    "sampy@attrs": {"units"},
                    "sampy@shape": (npoints,),
                    "sampz@attrs": {"units"},
                    "sampz@shape": (npoints,),
                },
                "rois_norm": {
                    "@attrs": {"NX_class", "auxiliary_signals", "axes", "signal"},
                    "roi1@shape": (npoints,),
                    "roi2@shape": (npoints,),
                    "sampy@attrs": {"units"},
                    "sampy@shape": (npoints,),
                    "sampz@attrs": {"units"},
                    "sampz@shape": (npoints,),
                },
            },
            "title@shape": (),
            "end_time@shape": (),
        }

        content[f"{scan}.1"] = scan_content
        for detector in range(ndetectors):
            scan_content["instrument"][f"mca{detector}"] = mca_detector
            add = {k.format(detector=detector): v for k, v in mca_meas.items()}
            scan_content["measurement"].update(add)

    return content



import logging
from typing import Any
from typing import Dict
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy

from ...io import hdf5
from ...math import pad
from .. import xrf_results
from ..math import eval_hdf5_expression
from ..math import format_expression_template

_logger = logging.getLogger(__name__)


def normalization_template(
    expression: Optional[str],
    counter_normalization_template: str,
    counter_name: Optional[str],
    detector_normalization_template: str,
    detector_name: Optional[str],
) -> str:

    factors = []

    if expression:
        try:
            expression.format()
        except IndexError:
            pass
        else:
            factors.append(expression)

    if detector_name:
        factors.append(
            format_expression_template(detector_normalization_template, detector_name)
        )

    if counter_name:
        factors.append(
            format_expression_template(counter_normalization_template, counter_name)
        )

    if not factors:
        raise ValueError(
            "specify at least one of these: 'expression_template', 'counter_name' or 'detector_name'"
        )

    return "*".join(factors)


def normalization_coefficient(
    weight_uri_root: str, normalization_expression: str
) -> numpy.ndarray:
    """
    :param weight_uri_root: HDF5 root group under which the normalization datasets can be found.
    :param normalization_expression: Arithmetic expression to calculated the normalization coefficient.
    """
    coefficient = eval_hdf5_expression(weight_uri_root, normalization_expression)
    if _logger.isEnabledFor(logging.DEBUG):
        _logger.debug(
            "Normalization coefficients [min=%f, median=%f, max=%s]",
            numpy.nanmin(coefficient),
            numpy.nanmedian(coefficient),
            numpy.nanmin(coefficient),
        )
    return coefficient


def normalization_coefficient_stack(
    weight_uri_roots: Sequence[str], normalization_expression: str
) -> numpy.ndarray:
    """
    :param weight_uri_roots: HDF5 root group under which the normalization datasets can be found.
    :param normalization_expression: Arithmetic expression to calculated the normalization coefficient.
    """
    coefficient = [
        eval_hdf5_expression(weight_uri_root, normalization_expression)
        for weight_uri_root in weight_uri_roots
    ]
    coefficient = numpy.stack(coefficient, axis=0)
    if _logger.isEnabledFor(logging.DEBUG):
        _logger.debug(
            "Normalization coefficients [min=%f, median=%f, max=%s]",
            numpy.nanmin(coefficient),
            numpy.nanmedian(coefficient),
            numpy.nanmin(coefficient),
        )
    return coefficient


def save_normalized_xrf_results(
    xrf_results_uri: str,
    coefficient: numpy.ndarray,
    output_root_uri: str,
    process_config: Dict[str, Any],
) -> str:
    """
    :param xrf_results_uri: HDF5 group that contains the "parameters", "uncertainties" and "massfractions" groups.
    :param coefficient: Coefficient with which to normalize the data.
    :param output_root_uri: Root HDF5 URL under which to save the results as NXprocess
    :returns: URI to normalized fit result group
    """
    parameters, uncertainties, massfractions = compute_normalized_xrf_results(
        xrf_results_uri, coefficient
    )
    return xrf_results.save_xrf_results(
        output_root_uri,
        "norm",
        process_config,
        parameters,
        uncertainties,
        massfractions,
    )


def compute_normalized_xrf_results(
    xrf_results_uri: str, coefficient: numpy.ndarray
) -> Tuple[
    Dict[str, numpy.ndarray], Dict[str, numpy.ndarray], Dict[str, numpy.ndarray]
]:
    r"""Compute the normalized peak areas, associated uncertainties and mass fractions.

    For elemental peak areas

    .. math::

        A(\mathrm{Fe}) = C \cdot A(\mathrm{Fe})

    For their uncertainties

    .. math::

        \sigma_{A}(\mathrm{Fe}) = C \cdot \sigma_{A}(\mathrm{Fe})  }

    :param xrf_results_uri: HDF5 group that contains the "parameters", "uncertainties" and "massfractions" groups.
    :param coefficient: Coefficient with which to normalize the data.
    :returns: normalized peak areas, associated uncertainties, normalized weight fractions
    """
    normalized_parameters = {}
    normalized_uncertainties = {}
    normalized_massfractions = {}

    fit_filename, fit_h5path = xrf_results_uri.split("::")

    with hdf5.FileReadAccess(fit_filename) as h5file:
        xrf_results_group = h5file[fit_h5path]
        assert hdf5.is_group(xrf_results_group)

        massfrac_group = xrf_results_group.get("massfractions", dict())

        # Peak areas and mass fractions
        param_group = xrf_results_group["parameters"]
        assert hdf5.is_group(param_group)
        for dset_name, dset in param_group.items():
            if not xrf_results.is_peak_area(dset):
                continue

            wparam_value = _normalize_dataset(dset[()], coefficient)
            normalized_parameters[dset_name] = wparam_value

            if dset_name not in massfrac_group:
                continue

            wmassfrac_value = _normalize_dataset(
                massfrac_group[dset_name][()], coefficient
            )
            normalized_massfractions[dset_name] = wmassfrac_value

        # Propagate error on peak areas
        uncertainties_group = xrf_results_group["uncertainties"]
        assert hdf5.is_group(uncertainties_group)
        for dset_name, dset in uncertainties_group.items():
            if not hdf5.is_dataset(dset):
                continue
            wsigma_value = _normalize_dataset(dset, coefficient)
            normalized_uncertainties[dset_name] = wsigma_value

    return normalized_parameters, normalized_uncertainties, normalized_massfractions


def _normalize_dataset(
    dset: hdf5.DatasetType, coefficients: Union[numpy.ndarray, float, int]
) -> numpy.ndarray:
    if numpy.isscalar(coefficients):
        return dset[()] * coefficients
    ndset = len(dset)
    ncoeff = len(coefficients)
    if ndset == ncoeff:
        return dset[()] * coefficients
    if ndset < ncoeff:
        return pad.pad_array(dset[()], ncoeff) * coefficients
    return dset[()] + pad.pad_array(coefficients, ndset)



from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import numpy

from ...io import hdf5


def read_position_suburis(
    bliss_scan_uri: str,
    position_suburis: Sequence[str],
    axes_units: Optional[Dict[str, str]] = None,
) -> Tuple[List[numpy.ndarray], List[str], List[Optional[str]]]:
    data = [
        get_position_data(bliss_scan_uri, position_suburi)
        for position_suburi in position_suburis
    ]
    positions, units = zip(*data)
    names = [[s for s in name.split("/") if s][-1] for name in position_suburis]
    if axes_units:
        units = [
            unit if unit else axes_units.get(name) for name, unit in zip(names, units)
        ]
    return positions, names, units


def get_position_data(
    bliss_scan_uri: str, position_suburi: str
) -> Tuple[numpy.ndarray, Optional[str]]:
    """Get position data with units from HDF5"""
    scan_filename, scan_h5path = hdf5.split_h5uri(bliss_scan_uri)

    with hdf5.FileReadAccess(scan_filename) as scan_file:
        scan_grp = scan_file[scan_h5path]
        pos_dataset = scan_grp[position_suburi]
        assert hdf5.is_dataset(pos_dataset)

        unit = pos_dataset.attrs.get("units")
        if not unit:
            posi = position_suburi.split("/")[-1]
            url = f"instrument/positioners_start/{posi}"
            if url in scan_grp:
                unit = scan_grp[url].attrs.get("units")

        return pos_dataset[()], unit


def get_scan_position_suburis(
    bliss_scan_uri: str, ignore_positioners: Optional[Sequence[str]] = None
) -> List[str]:
    """Get all scan sub-URI's for positioners which were scanned."""
    scan_filename, scan_h5path = hdf5.split_h5uri(bliss_scan_uri)

    with hdf5.FileReadAccess(scan_filename) as scan_file:
        scan_grp = scan_file[scan_h5path]
        positioners = scan_grp["instrument/positioners_start"]
        positioners = set(positioners)
        counters = set(scan_grp["measurement"])
        positioners &= counters
        if ignore_positioners:
            positioners -= set(ignore_positioners)
        # E.g. order ["sampz", "sampy"]
        # The first dimension in pymca and silx is plotted vertically
        return [f"measurement/{s}" for s in reversed(sorted(positioners))]



from typing import Sequence
from typing import Tuple

import numpy
from sklearn.neighbors import NearestNeighbors


def expanded_grid_coordinates(grid_axes: Sequence[numpy.ndarray]) -> numpy.ndarray:
    """
    Expand grid axes to coordinates of grid nodes.

    :param grid_axes: `nDim` arrays with shape `(N0,)`, `(N1,)`, ...
    :returns: shape `(N0*N1*... , Ndim)`
    """
    grid_coords = numpy.meshgrid(*grid_axes, indexing="ij")
    return numpy.vstack([g.ravel() for g in grid_coords]).T


def closest_point(
    coordinates1: numpy.ndarray, coordinates2: numpy.ndarray
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Find the closest points in `coordinates1` for each point in `coordinates2`.

    :param coordinates1: Coordinates with shape `(N1, Ndim)`
    :param coordinates2: Coordinates with shape `(N2, Ndim)`
    :returns: A tuple containing:
              - indices of the closest points in `coordinates2` for each point in `coordinates1` with shape `(N1,)`
              - distances to the closest points with shape `(N1,)`
    """
    nbrs = NearestNeighbors(n_neighbors=1, algorithm="kd_tree")
    nbrs.fit(coordinates2)
    distances, indices = nbrs.kneighbors(coordinates1)
    return indices.flatten(), distances.flatten()



import logging
import os
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import h5py
import numpy
from numpy.typing import DTypeLike
from pint import Quantity

from ..math.expression import eval_expression
from ..math.expression import expression_variables
from ..units import unit_registry
from ..units import units_as_str
from . import hdf5

logger = logging.getLogger(__name__)


def concatenate_bliss_scans(
    bliss_scan_uris: List[str],
    output_root_uri: str,
    virtual_axes: Optional[Dict[str, str]] = None,
    axes_units: Optional[Dict[str, str]] = None,
    start_var: str = "<",
    end_var: str = ">",
) -> str:
    """Concatenate Bliss scans in a virtual scan that looks exactly like a Bliss scan.

    This method cannot handle scans that are interrupted, except when it is the last scan to be concatenated.

    :param bliss_scan_uris: scans to concatenate
    :param virtual_axes: virtual motors. For example `{"sy": "<samy>+<sampy>/1000", "sz": "<samz>+<sampz>/1000"}`
    :param axes_units: axis units. For example `{"samy": "mm", "sampy": "um"}`
    :param start_var: marks the start of a variable name in `virtual_axes` expressions
    :param end_var: marks the end of a variable name in `virtual_axes` expressions
    :returns: URI to the virtual concatenated scan
    """
    if axes_units is None:
        axes_units = dict()

    out_filename, out_path = hdf5.split_h5uri(output_root_uri)
    if not out_path:
        out_path = "/1.1"
        output_root_uri = hdf5.join_h5url(out_filename, out_path)

    # Return when out_path already exists
    with h5py.File(out_filename, mode="a") as out_root:
        if out_path in out_root:
            logger.warning("%s already exists", output_root_uri)
            return output_root_uri

    # Compile data to concatenate
    in_scan = None
    datasets = None
    for bliss_scan_uri in bliss_scan_uris:
        logger.debug("Parse %s for concatenation", bliss_scan_uri)
        in_filename, in_path = hdf5.split_h5uri(bliss_scan_uri)
        with h5py.File(in_filename, mode="r", locking=False) as in_root:
            in_scani = in_root[in_path]
            if in_scan is None:
                in_scan, datasets = _parse_hdf5_group(in_scani)
            else:
                _append_H5Datasets(in_scani, datasets)

    logger.debug("Save concatenation in %s", output_root_uri)

    skip_datasets = [name for name, dataset in datasets.items() if dataset.skip]
    datasets = {name: dataset for name, dataset in datasets.items() if not dataset.skip}
    scalar_repeats = _trim_datasets(datasets)

    with h5py.File(out_filename, mode="a") as out_root:
        out_scan = out_root.create_group(out_path)
        _save_hdf5_group(out_scan, in_scan, scalar_repeats, skip_datasets)

    if virtual_axes:
        virtual_axis_datasets = _add_virtual_axes(
            in_scan, virtual_axes, axes_units, start_var=start_var, end_var=end_var
        )

        with h5py.File(out_filename, mode="r") as out_root:
            out_scan = out_root[out_path]
            for item in virtual_axis_datasets:
                _resolve_virtual_axis(item, out_scan, axes_units)

        with h5py.File(out_filename, mode="a") as out_root:
            out_scan = out_root[out_path]
            _save_hdf5_group(
                out_scan, in_scan, scalar_repeats, skip_datasets, skip_existing=True
            )

    return output_root_uri


@dataclass
class _H5SoftLink:
    path: str
    relative: bool


@dataclass
class _H5VirtualSource:
    filename: str
    data_path: str
    shape: Tuple[int]
    dtype: DTypeLike
    force_shape0: Optional[int] = None

    @property
    def layout_shape(self) -> Tuple[int]:
        if self.force_shape0 is None:
            return self.shape
        return (self.force_shape0,) + self.shape[1:]

    def get_layout_h5item(self, filename: str) -> h5py.VirtualSource:
        ext_filename = os.path.relpath(self.filename, os.path.dirname(filename))
        return h5py.VirtualSource(
            ext_filename, self.data_path, shape=self.shape, dtype=self.dtype
        )


@dataclass
class _H5Item:
    attrs: dict


@dataclass
class _H5Group(_H5Item):
    items: Dict[str, Union[_H5Item, _H5SoftLink]]


@dataclass
class _H5Dataset(_H5Item):
    value: Any


@dataclass
class _H5DatasetExpression(_H5Item):
    expression: str
    units: str
    start_var: str = "<"
    end_var: str = ">"
    result: Optional[Quantity] = None


@dataclass
class _H5ScalarDataset(_H5Item):
    values: List[numpy.number]
    skip: bool = False


@dataclass
class _VirtualDataset(_H5Item):
    sources: List[_H5VirtualSource]
    skip: bool = False

    @property
    def dtype(self) -> DTypeLike:
        return self.sources[0].dtype

    @property
    def shape(self) -> Tuple[int]:
        s0 = sum(source_scani.layout_shape[0] for source_scani in self.sources)
        sother = self.sources[0].layout_shape[1:]
        return (s0,) + sother

    @property
    def __len__(self) -> int:
        return self.shape[0]


def _parse_hdf5_group(
    parent: hdf5.GroupType, _name_prefix_strip: int = 0
) -> Tuple[_H5Group, Dict[str, Union[_VirtualDataset, _H5ScalarDataset]]]:
    if _name_prefix_strip <= 0:
        _name_prefix_strip = len(parent.name) + 1
    filename = parent.file.filename
    items = dict()
    group = _H5Group(attrs=dict(parent.attrs), items=items)
    datasets = dict()
    for name in parent:
        link = parent.get(name, getlink=True)
        if isinstance(link, h5py.HardLink):
            child = parent.get(name)
            if hdf5.is_dataset(child):
                if numpy.issubdtype(child.dtype, numpy.number):
                    if child.ndim == 0:
                        values = [child[()]]
                        dataset = _H5ScalarDataset(
                            attrs=dict(child.attrs), values=values
                        )
                    else:
                        sources = [
                            _H5VirtualSource(
                                filename=filename,
                                data_path=child.name,
                                shape=child.shape,
                                dtype=child.dtype,
                            )
                        ]
                        dataset = _VirtualDataset(
                            attrs=dict(child.attrs), sources=sources
                        )
                    key = child.name[_name_prefix_strip:]
                    datasets[key] = dataset
                    items[name] = dataset
                else:
                    value = child[()]
                    if isinstance(value, bytes):
                        value = value.decode()
                    items[name] = _H5Dataset(attrs=dict(child.attrs), value=value)
            elif hdf5.is_group(child):
                if child.attrs.get("NX_class", "") == "NXdata":
                    continue
                items[name], sub_H5Datasets = _parse_hdf5_group(
                    child, _name_prefix_strip=_name_prefix_strip
                )
                datasets.update(sub_H5Datasets)
            else:
                logger.warning(
                    f"ignore HDF5 item {parent.name}/{name} for concatenation ({type(child)})"
                )
        elif isinstance(link, h5py.SoftLink):
            if link.path.startswith("/"):
                items[name] = _H5SoftLink(
                    path=link.path[_name_prefix_strip:], relative=False
                )
            else:
                items[name] = _H5SoftLink(path=link.path, relative=True)
        else:
            logger.warning(
                f"ignore HDF5 link {parent.name}/{name} for concatenation ({type(link)})"
            )
    return group, datasets


def _append_H5Datasets(
    group: hdf5.GroupType, datasets: Dict[str, Union[_VirtualDataset, _H5ScalarDataset]]
) -> None:
    filename = group.file.filename
    for name, dataset in datasets.items():
        if dataset.skip:
            continue
        if name not in group:
            logger.warning(
                "Skip dataset %s/%s because not present in %s",
                group.name,
                name,
                group.file.filename,
            )
            dataset.skip = True
            continue
        child = group[name]
        if isinstance(dataset, _H5ScalarDataset):
            dataset.values.append(child[()])
        else:
            dataset.sources.append(
                _H5VirtualSource(
                    filename=filename,
                    data_path=child.name,
                    shape=child.shape,
                    dtype=child.dtype,
                )
            )


def _trim_datasets(
    datasets: Dict[str, Union[_VirtualDataset, _H5ScalarDataset]],
) -> List[int]:
    dataset_names, dataset_lengths = _dataset_lengths_per_scan(datasets)

    min_dataset_lengths = list()

    for scan_index, dataset_lengths_scani in enumerate(dataset_lengths):
        min_dataset_length_scani = min(dataset_lengths_scani)
        min_dataset_lengths.append(min_dataset_length_scani)

        for dataset_name in dataset_names:
            vsource_scani = datasets[dataset_name].sources[scan_index]
            vsource_scani.force_shape0 = min_dataset_length_scani

        # Even if you have know the logic of a dataset length, for example
        #
        #   dataset_length = (nfast+1) * nslow
        #
        # I would not know how to trim it so just trim to obtain an equal number of points.

    return min_dataset_lengths


def _dataset_lengths_per_scan(
    datasets: Dict[str, Union[_VirtualDataset, _H5ScalarDataset]],
) -> Tuple[List[str], List[List[int]]]:
    """
    :returns: dataset names with shape `(ndatasets,)` and dataset lengths with shape `(nscans, ndatasets)`
    """
    dataset_lengths = list()
    dataset_names = list()
    for name, dataset in datasets.items():
        if not isinstance(dataset, _VirtualDataset):
            continue
        dataset_names.append(name)
        if dataset_lengths:
            for source_scani, dataset_lengths_scani in zip(
                dataset.sources, dataset_lengths
            ):
                dataset_lengths_scani.append(source_scani.shape[0])
        else:
            dataset_lengths = [
                [source_scani.shape[0]] for source_scani in dataset.sources
            ]
    return dataset_names, dataset_lengths


def _save_hdf5_group(
    group: hdf5.GroupType,
    structure: _H5Group,
    scalar_repeats: List[int],
    skip_datasets: List[str],
    skip_existing: bool = False,
    _name_prefix: str = "",
) -> None:
    filename = group.file.filename
    group.attrs.update(structure.attrs)
    if not _name_prefix:
        _name_prefix = group.name
    for name, item in structure.items.items():
        skip_item = skip_existing and name in group

        if isinstance(item, _H5Group):
            if skip_item:
                subgroup = group[name]
            else:
                subgroup = group.create_group(name)
            _save_hdf5_group(
                subgroup,
                item,
                scalar_repeats,
                skip_datasets,
                skip_existing=skip_existing,
                _name_prefix=_name_prefix,
            )
            continue

        if skip_item:
            continue

        if isinstance(item, _H5SoftLink):
            if item.relative:
                group[name] = h5py.SoftLink(item.path)
            else:
                group[name] = h5py.SoftLink(f"{_name_prefix}/{item.path}")
        elif isinstance(item, _H5Dataset):
            group[name] = item.value
            group[name].attrs.update(item.attrs)
        elif isinstance(item, _H5ScalarDataset):
            if item.skip:
                continue
            if len(set(item.values)) > 1:
                # No longer a scalar dataset
                group[name] = numpy.repeat(item.values, scalar_repeats)
            else:
                group[name] = item.values[0]
            group[name].attrs.update(item.attrs)
        elif isinstance(item, _VirtualDataset):
            if item.skip:
                continue
            layout = h5py.VirtualLayout(shape=item.shape, dtype=item.dtype)
            start_index = 0
            for source_scani in item.sources:
                n = source_scani.layout_shape[0]
                vsource = source_scani.get_layout_h5item(filename)
                layout[start_index : start_index + n] = vsource[:n]
                start_index += n
            group.create_virtual_dataset(name, layout, fillvalue=numpy.nan)
            group[name].attrs.update(item.attrs)
        elif isinstance(item, _H5DatasetExpression):
            if item.result is None:
                continue
            group[name] = item.result.magnitude
            if not item.result.units.dimensionless:
                group[name].attrs["units"] = units_as_str(item.result.units)
        else:
            logger.debug(f"ignore HDF5 item {name} for saving ({type(item)})")


def _ensure_same_shape0(
    variables: Dict[str, numpy.ndarray], name_map: Dict[str, str]
) -> None:
    npoints = dict()
    for key, values in variables.items():
        if values.ndim >= 1:
            npoints[key] = len(values)
    if len(set(npoints.values())) <= 1:
        return
    npoints_min = min(npoints.values())
    for key, npoints_orig in npoints.items():
        logger.warning(
            "trim '%s' from %d points to %d", name_map[key], npoints_orig, npoints_min
        )
        variables[key] = variables[key][:npoints_min]


def _add_virtual_axes(
    group: _H5Group,
    virtual_axes: Dict[str, str],
    axes_units: Dict[str, str],
    start_var: str = "<",
    end_var: str = ">",
) -> List[_H5DatasetExpression]:
    instrument = group.items["instrument"]
    measurement = group.items["measurement"]
    positioners = instrument.items["positioners"]

    datasets = list()
    for motor_name, expression in virtual_axes.items():
        motor_value = _H5DatasetExpression(
            attrs=dict(),
            expression=expression,
            units=axes_units.get(motor_name, ""),
            start_var=start_var,
            end_var=end_var,
        )
        instrument.items[motor_name] = _H5Group(
            attrs={"NX_class": "NXpositioner"}, items={"value": motor_value}
        )
        measurement.items[motor_name] = _H5SoftLink(
            f"instrument/{motor_name}/value", relative=False
        )
        positioners.items[motor_name] = _H5SoftLink(
            f"instrument/{motor_name}/value", relative=False
        )
        datasets.append(motor_value)
    return datasets


def _resolve_virtual_axis(
    item: _H5DatasetExpression, top_group: hdf5.GroupType, axes_units: Dict[str, str]
) -> None:
    def get_data(name: str) -> Tuple[str, Quantity]:
        return _get_moving_values(
            top_group, name, axes_units.get(name, "") if axes_units else ""
        )

    expression, variables, name_map = expression_variables(
        item.expression, get_data, start_var=item.start_var, end_var=item.end_var
    )
    _ensure_same_shape0(variables, name_map)
    quantity = eval_expression(expression, variables)
    if item.units:
        quantity = quantity.to(item.units)
    item.result = quantity


def _get_moving_values(
    scan: hdf5.GroupType, name: str, default_units: str
) -> Tuple[str, Quantity]:
    measurement = scan["measurement"]
    if name in measurement:
        full_name = f"{measurement.name}/{name}"
        values = _get_value_with_units(measurement[name], default_units)
        return full_name, values
    instrument = scan["instrument"]
    if name in instrument:
        group = instrument[name]
        if "value" in group:
            # NXpositioner
            full_name = f"{group.name}/value"
            values = _get_value_with_units(group["value"], default_units)
            return full_name, values
        else:
            # NXdetector
            full_name = f"{group.name}/data"
            values = _get_value_with_units(group["data"], default_units)
            return full_name, values
    positioners = instrument["positioners"]
    if name in positioners:
        full_name = f"{positioners.name}/{name}"
        values = _get_value_with_units(positioners[name], default_units)
        return full_name, values
    raise RuntimeError(f"'{name}' is neither a detector nor a positioner")


def _get_value_with_units(dset: hdf5.DatasetType, default_units: str) -> Quantity:
    units = dset.attrs.get("units", default_units)
    ureg = unit_registry()
    return dset[()] * ureg.parse_units(units)



import re
from typing import List
from typing import Tuple


def integers_serializer(value: List[int]) -> str:
    return ", ".join([str(n) for n in value])


def integers_deserializer(value: str) -> List[int]:
    return [int(n) for n in value.replace(" ", "").split(",")]


def strings_serializer(value: List[str]) -> str:
    return ", ".join([s for s in value])


def strings_deserializer(value: str) -> List[str]:
    return [s for s in value.replace(" ", "").split(",")]


def shape_serializer(value: Tuple[int]) -> str:
    return ", ".join([str(n) for n in value])


def shape_deserializer(value: str) -> List[int]:
    return tuple(int(n) for n in value.replace(" ", "").split(","))


def rois_serializer(value: List[List[int]]) -> str:
    rois = [f"[{','.join([str(n) for n in tpl])}]" for tpl in value]
    return ", ".join(rois)


def rois_deserializer(value: str) -> List[List[int]]:
    pattern = r"\[[0-9,]+\]"
    itrois = re.finditer(pattern, value.replace(" ", ""))
    return [[int(n) for n in roi[1:-1].split(",")] for roi in itrois]



import os
from contextlib import contextmanager
from typing import Optional

from ewoksdata.data.hdf5 import h5exists
from silx.io.url import DataUrl

from .abstract import AbstractOutputBuffer


class PyMcaOutputBuffer(AbstractOutputBuffer):
    """This is the output buffer of PyMca with an internal output handler."""

    def __init__(
        self,
        output_root_uri: str,
        diagnostics: bool = False,
        figuresofmerit: bool = False,
        **open_options,
    ):
        url = DataUrl(output_root_uri)
        filename = str(url.file_path())
        h5path = [s for s in url.data_path().split("/") if s]
        nh5path = len(h5path)
        if nh5path == 0:
            h5path = "results", "fit"
        elif nh5path == 1:
            h5path.append("fit")
        else:
            h5path = h5path[:2]
        self._file_entry, self._file_process = h5path
        self._output_dir = os.path.dirname(filename)
        self._output_root = os.path.splitext(os.path.basename(filename))[0]
        super().__init__(
            outputDir=self._output_dir,
            outputRoot=self._output_root,
            fileEntry=self._file_entry,
            fileProcess=self._file_process,
            saveResiduals=diagnostics,
            saveFit=diagnostics,
            saveData=diagnostics,
            diagnostics=diagnostics,
            saveFOM=figuresofmerit,
        )
        self._open_options = open_options  # e.g. retry arguments
        open_options["mode"] = "a"
        self._newgroup = False
        self._xrf_results_uri = None

    @property
    def xrf_results_uri(self) -> Optional[str]:
        return self._xrf_results_uri

    @property
    def already_existed(self) -> bool:
        return self._newgroup

    @contextmanager
    def saveContext(self, **kw):
        filename = os.path.join(self._output_dir, self._output_root + ".h5")
        h5path = f"{self._file_entry}/{self._file_process}"
        self._newgroup = not h5exists(filename, h5path, **self._open_options)
        self._xrf_results_uri = f"{filename}::/{h5path}/results"
        with super().saveContext(**kw) as ctx:
            yield ctx



from pathlib import Path
from typing import Any
from typing import Tuple

import numpy
import pytest
from silx.io import dictdump
from silx.io import h5py_utils

from ..io.hdf5 import split_h5uri
from ..tasks import hdf5_utils


@pytest.fixture()
def dummy_bliss_scan(tmp_path: Path) -> Tuple[str, dict]:
    filename = tmp_path / "raw_data.py"

    samy = numpy.arange(0, 2 * numpy.pi, 360)
    diode1 = numpy.sin(numpy.arange(0, 2 * numpy.pi, 360))

    data = {
        "@NX_class": "NXroot",
        "1.1": {
            "@NX_class": "NXentry",
            "instrument": {
                "@NX_class": "NXinstrument",
                "diode1": {
                    "@NX_class": "NXdetector",
                    "data": diode1,
                },
                "samy": {
                    "@NX_class": "NXpositioner",
                    "value": samy,
                },
                "positioners": {"@NX_class": "NXCollection", ">samy": "../samy/value"},
            },
            "sample": {
                "@NX_class": "NXsample",
            },
            "measurement": {
                "@NX_class": "NXcollection",
                ">diode1": "../instrument/diode1/data",
                ">samy": "../instrument/samy/value",
            },
        },
    }

    dictdump.dicttonx(data, filename, "/")

    expected = {
        "@NX_class": "NXroot",
        "1.1": {
            "@NX_class": "NXentry",
            "instrument": {
                "@NX_class": "NXinstrument",
                "diode1": {"@NX_class": "NXdetector", "data": diode1},
                "samy": {"@NX_class": "NXpositioner", "value": samy},
                "positioners": {"@NX_class": "NXCollection", "samy": samy},
            },
            "sample": {
                "@NX_class": "NXsample",
            },
            "measurement": {
                "@NX_class": "NXcollection",
                "diode1": diode1,
                "samy": samy,
            },
        },
    }

    return f"{filename}::/1.1", expected


def test_link_bliss_scan(dummy_bliss_scan: Tuple[str, dict], tmp_path: Path):
    bliss_scan_url, expected = dummy_bliss_scan

    _, data_h5path = split_h5uri(bliss_scan_url)
    entry_name = [s for s in data_h5path.split("/") if s][0]

    filename = tmp_path / "processed_data.py"

    with h5py_utils.File(filename, mode="a") as root:
        data = {
            "@NX_class": "NXroot",
            entry_name: {
                "@NX_class": "NXentry",
                "fit": {"@NX_class": "NXprocess"},
            },
        }
        expected[entry_name]["fit"] = data[entry_name]["fit"]
        dictdump.dicttonx(data, root)

        hdf5_utils.link_bliss_scan(root[entry_name], bliss_scan_url, retry_timeout=0)

    _assert_content(filename, expected)


def test_link_bliss_scan_same_file(dummy_bliss_scan: Tuple[str, dict]):
    bliss_scan_url, expected = dummy_bliss_scan

    filename, data_h5path = split_h5uri(bliss_scan_url)
    entry_name = [s for s in data_h5path.split("/") if s][0]

    with h5py_utils.File(filename, mode="a") as root:
        data = {
            entry_name: {
                "fit": {"@NX_class": "NXprocess"},
            },
        }
        expected[entry_name]["fit"] = data[entry_name]["fit"]
        dictdump.dicttonx(data, root)

        hdf5_utils.link_bliss_scan(root[entry_name], bliss_scan_url, retry_timeout=0)

    _assert_content(filename, expected)


def _assert_content(filename: Path, expected: dict) -> None:
    actual = dictdump.nxtodict(
        filename, include_attributes=True, dereference_links=True
    )
    actual = _convert_arrays(actual)
    expected = _convert_arrays(expected)
    assert actual == expected


def _convert_arrays(obj: Any) -> Any:
    if isinstance(obj, numpy.ndarray):
        return obj.tolist()
    elif isinstance(obj, dict):
        return {k: _convert_arrays(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_convert_arrays(v) for v in obj]
    else:
        return obj



import h5py
import pytest

from ..tasks import positioner_utils


@pytest.fixture(scope="module")
def example_bliss_scan_uri(tmpdir_factory):
    filename = str(tmpdir_factory.mktemp("test_positioners").join("data.h5"))
    with h5py.File(filename, mode="w") as root:
        dset = root.create_dataset(
            "/1.1/instrument/positioners_start/energy", data=10.0
        )
        dset.attrs["units"] = "keV"
        dset = root.create_dataset(
            "/1.1/instrument/positioners_start/energy_und",
            data=[10.0, 10.0, 10.0000001],
        )
        dset = root.create_dataset("/1.1/instrument/other/name", data=11.0)
    return f"{filename}::/1.1"


def test_get_energy(example_bliss_scan_uri):
    value = positioner_utils.get_primary_beam_energy_value(example_bliss_scan_uri)
    assert value is None
    value = positioner_utils.get_primary_beam_energy_value(
        example_bliss_scan_uri, search_on_units=True
    )
    assert value == 10.0
    value = positioner_utils.get_primary_beam_energy_value(
        example_bliss_scan_uri, energy_name="energy"
    )
    assert value == 10.0
    value = positioner_utils.get_primary_beam_energy_value(
        example_bliss_scan_uri, energy_name="energy_und"
    )
    assert value == 10.0
    value = positioner_utils.get_primary_beam_energy_value(
        example_bliss_scan_uri,
        energy_name="energy",
        energy_uri_template="instrument/positioners_start/{}",
    )
    assert value == 10.0
    value = positioner_utils.get_primary_beam_energy_value(
        example_bliss_scan_uri,
        energy_name="name",
        energy_uri_template="instrument/other/{}",
    )
    assert value == 11.0


def test_get_energy_suburi(example_bliss_scan_uri):
    value = positioner_utils.get_primary_beam_energy_suburi(example_bliss_scan_uri)
    assert value == "instrument/positioners_start/energy"



import re
from typing import Any
from typing import Callable
from typing import Dict
from typing import Tuple

import numpy

from . import pad


def expression_variables(
    expression: str,
    get_data: Callable[[str], Tuple[str, Any]],
    start_var: str = "<",
    end_var: str = ">",
) -> Tuple[str, Dict[str, Any], Dict[str, str]]:
    """Return the variables and associated expression to be evaluated.

    :param expression: for example `"<name A> * <name B>"`
    :param get_data: takes a variable name as input and returns its full name and value
    :param start_var: marks the start of a variable name
    :param end_var: marks the end of a variable name
    :returns: expression, variables and name map. For example expression `"data0 * data1"`,
              variables `{"data0":..., "data1":...}` and name map `{"data0":"name A", "data1":"name B"}`.
    """
    pattern = rf"{re.escape(start_var)}([^{re.escape(end_var)}]+){re.escape(end_var)}"

    variables = {}
    name_map = {}
    for i, variable_name in enumerate(re.findall(pattern, expression)):
        new_name = f"data{i}"
        name_map[new_name], variables[new_name] = get_data(variable_name)
        expression = expression.replace(
            f"{start_var}{variable_name}{end_var}", new_name
        )

    max_len = pad.pad_length(variables)
    if max_len is not None:
        pad.pad_arrays(max_len, variables)

    return expression, variables, name_map


def eval_expression(expression: str, variables: Dict[str, Any]) -> Any:
    """Evaluate an arithmetic expression with python and numpy arithmetic.

    :param expression: arithmetic expression where datasets are define as
                       `data1` where `data1`
                       must be a key in `variables`.
    :param variables: variables to be used in the expression
    """
    globals = {"__builtins__": {"len": len, "sum": sum}, "np": numpy, "numpy": numpy}
    return eval(expression, globals, variables)



from typing import Any

from ewokscore.model import BaseInputModel
from pydantic import ValidationInfo
from pydantic import field_validator


class FluoInputs(BaseInputModel):

    @field_validator("*", mode="before")
    @classmethod
    def none_to_default(cls, value: Any, info: ValidationInfo) -> Any:
        if value is not None:
            return value

        field = cls.model_fields[info.field_name]

        if field.default is not None:
            return field.default

        if field.default_factory is not None:
            return field.default_factory()

        return value



import logging
import os
import sys
import tempfile
import traceback
from contextlib import contextmanager
from functools import cached_property
from types import MappingProxyType
from typing import Generator
from typing import Iterator
from typing import Optional
from typing import Tuple
from typing import Union

import h5py
import numpy
from AnyQt import QtCore
from AnyQt import QtWidgets
from AnyQt.QtCore import Qt
from silx.app.view.Viewer import Viewer as SilxViewer
from silx.io import commonh5
from silx.io import h5py_utils

from ..io.hdf5 import FileReadAccess

_logger = logging.getLogger(__name__)


class DataViewer(SilxViewer):
    """Browse data from files supported by silx.

    To create the widget

    .. code: python

        viewer = DataViewer(parent)
        viewer.setVisible(True)
        parent.layout().addWidget(viewer)

    To close and refresh files

    .. code: python

        viewer.updateFile("/path/to/file1.h5")
        viewer.updateFile("/path/to/file2.h5")
        viewer.closeFile("/path/to/file1.h5")

    To close all files

    .. code: python

        viewer.closeAll()
    """

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        if parent is not None:
            self.setWindowFlags(Qt.Widget)

        # We cannot handle file opening in the viewer
        # because files are also opened when synchronizing
        model = self._findHdf5TreeModel()
        model.insertFile = _insertFile.__get__(model, type(model))

    def close(self):
        if self.parent():
            self.parent().close()
        else:
            self.close()

    def closeFile(self, filename: str) -> None:
        """When the file is opened, close it."""
        _, h5file = self._getFileObject(filename)
        if h5file is None:
            return
        model = self._findHdf5TreeModel()
        model.removeH5pyObject(h5file)

    def updateFile(self, filename, **file_open_options):
        """When the file exists, append when not already appended and refresh view."""
        # Remove when non existing
        if not os.path.exists(filename):
            self.closeFile(filename)
            return

        # Append when missing
        index, h5file = self._getFileObject(filename)
        if h5file is None:
            self.appendFile(filename, **file_open_options)
            index, h5file = self._getFileObject(filename)

        # Select the file (TODO: errors)
        selection_model = self._treeview.selectionModel()
        selection_model.clearSelection()
        self._treeview.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        selection_model.select(
            index, QtCore.QItemSelectionModel.Select | QtCore.QItemSelectionModel.Rows
        )
        self._treeview.setCurrentIndex(index)

        # Refresh the current selection
        self._Viewer__refreshAction.trigger()

    @property
    def _treeview(self):
        return self._Viewer__treeview

    def _findHdf5TreeModel(self):
        return self._treeview.findHdf5TreeModel()

    def _iterH5FileObjects(
        self,
    ) -> Iterator[Tuple[QtCore.QModelIndex, "ViewerFile"]]:
        model = self._findHdf5TreeModel()
        root_index = QtCore.QModelIndex()
        root = model.nodeFromIndex(root_index)
        for row in range(root.childCount()):
            hdf5item = root.child(row)
            index = model.index(row, 0, root_index)
            yield index, hdf5item.obj

    def _getFileObject(
        self, filename
    ) -> Tuple[Optional[QtCore.QModelIndex], Optional["ViewerFile"]]:
        filename = os.path.normpath(os.path.abspath(filename))
        for index, h5file in self._iterH5FileObjects():
            filename2 = os.path.normpath(os.path.abspath(h5file.filename))
            if filename2 == filename:
                return index, h5file
        return None, None


def _insertFile(self, filename, row=-1):
    """Open the file with the viewer proxy that does not keep the file opened."""
    try:
        h5file = ViewerFile(filename)
        if self._Hdf5TreeModel__ownFiles:
            self._Hdf5TreeModel__openedFiles.append(h5file)
        self.sigH5pyObjectLoaded.emit(h5file, filename)
        self.insertH5pyObject(h5file, row=row, filename=filename)
    except IOError:
        _logger.debug("File '%s' can't be read.", filename, exc_info=True)
        raise


class ViewerDataset(commonh5.Dataset):
    """Proxy to a HDF5 dataset that does not keep the file opened."""

    def __init__(self, name: str, parent: Union["ViewerFile", "ViewerGroup"]):
        super().__init__(name, None, parent=parent, attrs=None)

    @contextmanager
    def _h5dataset(self) -> Generator[h5py.Dataset, None, None]:
        with self.file._h5open() as h5file:
            yield h5file[self.name]

    def _get_h5attribute(self, attr: str):
        with self._h5dataset() as h5dataset:
            return getattr(h5dataset, attr)

    dtype = cached_property(lambda self: self._get_h5attribute("dtype"))
    shape = cached_property(lambda self: self._get_h5attribute("shape"))
    size = cached_property(lambda self: self._get_h5attribute("size"))
    ndim = cached_property(lambda self: self._get_h5attribute("ndim"))
    compression = cached_property(lambda self: self._get_h5attribute("compression"))
    compression_opts = cached_property(
        lambda self: self._get_h5attribute("compression_opts")
    )
    chunks = cached_property(lambda self: self._get_h5attribute("chunks"))
    is_virtual = cached_property(lambda self: self._get_h5attribute("is_virtual"))
    virtual_sources = cached_property(
        lambda self: self._get_h5attribute("virtual_sources")
    )
    external = cached_property(lambda self: self._get_h5attribute("external"))

    def __len__(self):
        with self._h5dataset() as h5dataset:
            return len(h5dataset)

    def __getitem__(self, item):
        with self._h5dataset() as h5dataset:
            return h5dataset[item]

    def __iter__(self):
        return self[()].__iter__()

    def __bool__(self):
        with self._h5dataset() as h5dataset:
            return bool(h5dataset)

    def __getattr__(self, item):
        """Proxy to underlying numpy array methods.

        Called for example when doing `numpy.array(dataset)`.
        """
        data = self[()]
        if hasattr(data, item):
            return getattr(data, item)

        raise AttributeError("Dataset has no attribute %s" % item)

    @cached_property
    def attrs(self):
        with self._h5dataset() as h5dataset:
            return MappingProxyType(dict(h5dataset.attrs))

    @property
    def value(self):
        raise NotImplementedError()  # Should not be used: h5py v2 property

    def _get_data(self):
        # All method calling this method in the base class should be overridden
        stack_trace = traceback.format_stack()
        _logger.warning(
            f"ViewerDataset._get_data should not be called\nStack trace:\n{''.join(stack_trace)}"
        )
        return self[()]


class ViewerGroup(commonh5.Group):
    """Proxy to a HDF5 group that does not keep the file opened."""

    def __init__(self, name, parent):
        with parent.file._h5open() as h5file:
            full_name = f"{parent.name}/{name}"
            h5group = h5file[full_name]
            attrs = dict(h5group.attrs)
            super().__init__(name, parent=parent, attrs=attrs)
            _add_nodes(self, h5group)


class ViewerFile(commonh5.File):
    """Proxy to a HDF5 file that does not keep the file opened."""

    def __init__(self, name: str, **file_open_options):
        self._file_open_options = file_open_options
        with FileReadAccess(name, **self._file_open_options) as h5file:
            h5group = h5file["/"]  # for order preservation
            attrs = dict(h5group.attrs)
            super().__init__(name=name, mode="r", attrs=attrs)
            _add_nodes(self, h5group)

    @contextmanager
    def _h5open(self) -> Generator[h5py.File, None, None]:
        with FileReadAccess(self.filename, **self._file_open_options) as h5file:
            yield h5file


def _add_nodes(
    commongroup: Union[ViewerFile, ViewerGroup], h5item: Union[h5py.Group, h5py.File]
):
    for base_name in h5item:
        if h5item.get(base_name, default=None, getclass=True) is h5py.Group:
            commongroup.add_node(ViewerGroup(base_name, commongroup))
        else:
            commongroup.add_node(ViewerDataset(base_name, commongroup))


def generate_example_data(name):
    filename = os.path.join(tempfile.gettempdir(), name)
    with h5py_utils.File(filename, mode="w") as nxroot:
        nxroot.attrs["NX_class"] = "NXroot"
        nxroot.attrs["creator"] = "test"
        nxroot.attrs["default"] = "2.1"

        # 1D data
        nxentry = nxroot.create_group("1.1")
        nxentry.attrs["NX_class"] = "NXroot"
        nxentry.attrs["default"] = "plot"
        nxentry["title"] = "ascan samx 0 3 3 0.1"

        measurement = nxentry.create_group("measurement")
        measurement.attrs["NX_class"] = "NXcollection"
        measurement["samx"] = [0, 1, 2, 3]
        measurement["diode1"] = [0, -1, -2, -3]
        measurement["diode2"] = [3, 1, 2, 0]

        nxdata = nxentry.create_group("plot")
        nxdata.attrs["NX_class"] = "NXdata"
        nxdata.attrs["signal"] = "diode1"
        nxdata.attrs["auxiliary_signals"] = ["diode2"]
        nxdata.attrs["axes"] = ["samx"]
        nxdata["samx"] = h5py.SoftLink(measurement["samx"].name)
        nxdata["diode1"] = h5py.SoftLink(measurement["diode1"].name)
        nxdata["diode2"] = h5py.SoftLink(measurement["diode2"].name)

        # 2D data
        nxentry = nxroot.create_group("2.1")
        nxentry.attrs["NX_class"] = "NXroot"
        nxentry.attrs["default"] = "plot"
        nxentry["title"] = "amesh samx 0 3 3 samy 0 4 4 0.1"

        measurement = nxentry.create_group("measurement")
        measurement.attrs["NX_class"] = "NXcollection"
        measurement["samx"] = [0, 0.1, 0.2, 0.3]
        measurement["samy"] = [0, 0.1, 0.2, 0.3, 0.4]
        measurement["diode1"] = numpy.zeros((4, 5))
        measurement["diode2"] = numpy.ones((4, 5))

        nxdata = nxentry.create_group("plot")
        nxdata.attrs["NX_class"] = "NXdata"
        nxdata.attrs["signal"] = "diode1"
        nxdata.attrs["auxiliary_signals"] = ["diode2"]
        nxdata.attrs["axes"] = ["samx", "samy"]
        nxdata["samx"] = h5py.SoftLink(measurement["samx"].name)
        nxdata["samx_indices"] = 0
        nxdata["samy"] = h5py.SoftLink(measurement["samy"].name)
        nxdata["samy_indices"] = 1
        nxdata["diode1"] = h5py.SoftLink(measurement["diode1"].name)
        nxdata["diode2"] = h5py.SoftLink(measurement["diode2"].name)

    return filename


def main(argv=None) -> Optional[int]:
    import argparse

    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(
        description="Load HDF5 files and display their structure."
    )
    parser.add_argument("files", nargs="*", help="List of HDF5 files to load.")
    parser.add_argument(
        "--example", action="store_true", help="Generate and open example data."
    )
    args = parser.parse_args(args=argv[1:])

    files = args.files
    if args.example:
        files.insert(0, generate_example_data("data_viewer_example.h5"))

    app = QtWidgets.QApplication(argv)
    dataviewer = DataViewer()
    dataviewer.resize(1300, 500)
    dataviewer._Viewer__splitter.setSizes([500, 800])
    for filename in files:
        dataviewer.appendFile(filename)
    dataviewer.show()
    return app.exec_()


if __name__ == "__main__":
    sys.exit(main())



import logging
from typing import Optional

from ..model import PyMcaXrfConfiguration

logger = logging.getLogger(__name__)


def set_fit_weights(
    configuration: PyMcaXrfConfiguration, quantification: Optional[dict]
):
    if configuration.fit.fitweight:
        if quantification:
            logger.warning(
                "Fit weights enabled. May reduce fit quality for low-count spectra, "
                "but uncertainties will remain correct."
            )
        else:
            logger.warning(
                "Disabling fit weights because not quantification is done. "
                "Fit quality improves for low-count spectra, but uncertainties will be incorrect."
            )
            configuration.fit.fitweight = 0



import logging

from ..model import PyMcaXrfConfiguration

logger = logging.getLogger(__name__)


def force_fast_fitting(configuration: PyMcaXrfConfiguration):
    fit = configuration.fit

    if fit.linearfitflag != 1:
        fit.linearfitflag = 1
        logger.warning("PyMca configuration: force linear fitting")

    if fit.strategyflag:
        fit.strategyflag = 0
        logger.warning("PyMca configuration: disable fit strategy")

    if fit.stripflag and fit.stripalgorithm != 1:
        logger.warning(
            "PyMca configuration: disable background stripping (use SNIP if you want to take background into account)"
        )
        fit.stripflag = 0



from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from .nexus import save_as_bliss_scan
from .spec import SpecFile


def spec_to_bliss(
    spec_filename: str,
    bliss_filename: str,
    scans: Union[int, Sequence[int], None] = None,
    subscans: Union[int, Sequence[int], None] = None,
    mode: str = "a",
    **openargs,
) -> None:
    include_scans = _parse_scan_arguments(scans, subscans)
    with SpecFile(spec_filename) as specfile:
        if include_scans is None:
            include_scans = specfile.get_scans()
        for scannb, subscan in include_scans:
            n = specfile.get_number_of_subsubscans(scannb, subscan)
            for i, (scannb, subscan) in enumerate(zip([scannb] * n, [subscan] * n)):
                subsubscan = i + 1
                osubscan = subscan + i
                save_as_bliss_scan(
                    bliss_filename,
                    f"{scannb}.{osubscan}",
                    specfile.iter_data(scannb, subscan=subscan, subsubscan=subsubscan),
                    positioners=specfile.get_positioners(scannb, subscan=subscan),
                    title=specfile.get_title(scannb, subscan=subscan),
                    mode=mode,
                    **openargs,
                )
                if mode == "w":
                    mode = "a"


def _parse_scan_arguments(
    scans: Union[int, Sequence[int], None], subscans: Union[int, Sequence[int], None]
) -> Optional[List[Tuple[int, int]]]:
    if scans is None:
        return None
    if subscans is None:
        subscans = 1
    if not isinstance(scans, Sequence):
        scans = [scans]
    if not isinstance(subscans, Sequence):
        subscans = [subscans]
    if len(scans) == 1:
        scans = scans * len(subscans)
    if len(subscans) == 1:
        subscans = subscans * len(scans)
    if len(scans) != len(subscans):
        raise ValueError("number of 'scans' and 'subscans' is not equal")
    return list(zip(scans, subscans))



import datetime
import logging
import os
from contextlib import ExitStack
from contextlib import contextmanager
from typing import Iterator
from typing import Mapping
from typing import Optional

import h5py
from ewoksdata.data.hdf5.dataset_writer import DatasetWriter
from silx.io import h5py_utils

from . import hdf5
from .types import ScanData

logger = logging.getLogger(__name__)


def save_as_bliss_scan(
    filename: str,
    entry_name: str,
    scan_data: Iterator[ScanData],
    positioners: Optional[Mapping[str, float]] = None,
    title: Optional[str] = None,
    **openoptions,
):
    openoptions.setdefault("mode", "a")
    os.makedirs(os.path.abspath(os.path.dirname(filename)), exist_ok=True)
    with h5py_utils.File(filename, **openoptions) as f:
        if entry_name in f:
            logger.warning("%s::/%s already exists", filename, entry_name)
            return

        with ExitStack() as stack:
            entry = stack.enter_context(_writer_context(f, entry_name))
            instrument = entry.create_group("instrument")
            instrument.attrs["NX_class"] = "NXinstrument"
            measurement = entry.create_group("measurement")
            measurement.attrs["NX_class"] = "NXcollection"

            if title:
                entry["title"] = title

            if positioners:
                collection = instrument.create_group("positioners_start")
                collection.attrs["NX_class"] = "NXcollection"
                for k, v in positioners.items():
                    collection[k] = v
                positioners = dict(positioners)

            writers = dict()
            for data in scan_data:
                detector = instrument.require_group(data.group)
                if data.detector_type == "positioner":
                    detector.attrs["NX_class"] = "NXpositioner"
                    if positioners:
                        positioners[data.group] = data.data
                else:
                    detector.attrs["NX_class"] = "NXdetector"
                    if "type" not in detector and data.detector_type:
                        detector["type"] = data.detector_type

                key = data.group, data.name
                writer = writers.get(key)
                if writer is None:
                    writer = stack.enter_context(DatasetWriter(detector, data.name))
                    writers[key] = writer
                    if data.local_alias:
                        detector[data.local_alias] = h5py.SoftLink(writer.dataset_name)
                    if data.global_alias and data.global_alias not in measurement:
                        measurement[data.global_alias] = h5py.SoftLink(
                            writer.dataset_name
                        )

                writer.add_points(data.data)

        if positioners:
            collection = instrument.create_group("positioners")
            collection.attrs["NX_class"] = "NXcollection"
            for k, v in positioners.items():
                collection[k] = v


@contextmanager
def _writer_context(root: hdf5.FileType, entry_name: str) -> Iterator[h5py.Group]:
    root.attrs["NX_class"] = "NXroot"
    root.attrs["creator"] = "ewoksfluo"
    entry = root.create_group(entry_name)
    try:
        entry.attrs["NX_class"] = "NXentry"
        entry.attrs["start_time"] = _timestamp()
        writer = entry.create_group("writer")
        writer.attrs["NX_class"] = "NXnote"
        writer["status"] = "STARTING"
        yield entry
    except Exception:
        writer["status"][()] = "FAILED"
        raise
    else:
        writer["status"][()] = "SUCCEEDED"
    finally:
        entry["end_time"] = _timestamp()


def _timestamp() -> str:
    return datetime.datetime.now().astimezone().isoformat()



import pytest

from ..math import rounding


def test_round_to_significant():
    for m in [1, -1, 1e-3, 1e3]:
        assert rounding.round_to_significant(m * 99.99) == m * 100
        assert rounding.round_to_significant(m * 100.1) == m * 100
        assert rounding.round_to_significant(m * 100.9) == m * 101
        assert rounding.round_to_significant(m * 101) == m * 101
        assert rounding.round_to_significant(m * 1001) == m * 1000


def test_round_range():
    start, stop, step = 0, 1, 0.1
    start, stop, nbins = rounding.round_range(start, stop, step)
    assert start == 0
    assert stop == 1
    assert nbins == 10

    start, stop, step = 0, 1, 0.101
    start, stop, nbins = rounding.round_range(start, stop, step)
    assert start == pytest.approx(-0.005)
    assert stop == pytest.approx(1.005)
    assert nbins == 10

    start, stop, step = 0, 1, 0.099
    start, stop, nbins = rounding.round_range(start, stop, step)
    assert start == pytest.approx(0.005)
    assert stop == pytest.approx(0.995)
    assert nbins == 10



from typing import Any
from typing import Tuple

import h5py

from ..math.expression import eval_expression
from ..math.expression import expression_variables
from ..tasks.math import eval_hdf5_expression


def test_eval_expression():
    expression = "np.mean([1,a,3,b])"
    variables = {"a": 2, "b": 4}
    eval_expression(expression, variables) == 2.5


def test_expression_variables():
    data = {"name A": 1, "name B": 2}

    def get_data(name: str) -> Tuple[str, Any]:
        return name, data[name]

    expression, variables, name_map = expression_variables(
        "<name A> * <name B>", get_data
    )
    assert expression == "data0 * data1"
    assert variables == {"data0": 1, "data1": 2}
    assert name_map == {"data0": "name A", "data1": "name B"}


def test_expression_variables_nondefault_brackets():
    data = {"name A": 1, "name B": 2}

    def get_data(name: str) -> Tuple[str, Any]:
        return name, data[name]

    expression, variables, name_map = expression_variables(
        "{name A} * {name B}", get_data, start_var="{", end_var="}"
    )
    assert expression == "data0 * data1"
    assert variables == {"data0": 1, "data1": 2}
    assert name_map == {"data0": "name A", "data1": "name B"}


def test_eval_hdf5_expression(tmp_path):
    filename = str(tmp_path / "data.h5")
    with h5py.File(filename, mode="w") as f:
        f["group/a"] = 2
        f["b"] = 4

    expression = "np.mean([1,<group/a>,3,<b>])"

    eval_hdf5_expression(filename, expression) == 2.5

    filename = str(tmp_path / "data.h5")
    with h5py.File(filename, mode="w") as f:
        f["root/group/a"] = 2
        f["root/b"] = 4

    expression = "np.mean([1,<group/a>,3,<b>])"

    eval_hdf5_expression(filename + "::/root", expression) == 2.5



import numpy


def expand_data(
    axis_indices,
    data: numpy.ndarray,
    axis_size: int,
    axis: int = 0,
    fill_value=numpy.nan,
) -> numpy.ndarray:
    """Expand ``data`` defined on a subset of ``axis_indices`` into an
    array of length ``axis_size`` along the specified ``axis``.
    """
    shape = list(data.shape)
    shape[axis] = axis_size
    data_full = numpy.full(shape, fill_value, dtype=data.dtype)
    indexer = [slice(None)] * data.ndim
    indexer[axis] = axis_indices
    data_full[tuple(indexer)] = data
    return data_full



from typing import Optional

import numpy


def monitor_signal(
    expo_time: float,
    npoints: int,
    max_decay: float = 0.8,
    number_of_injections: int = 2,
    seed: Optional[int] = None,
):
    """Monitor signal between :math:``max_decay`` and :math:``1`` with
    a number of injections defined by :math:``number_of_injections``.
    """
    # All points in time at which we want the calculate the monitor signal.
    time = numpy.arange(npoints) * expo_time
    tmin = 0
    tmax = time[-1]

    # Fixed time distance between two injections.
    injection_period = (tmax - tmin) / number_of_injections

    # Fix intensity decay after each injection.
    decay = -numpy.log(max_decay) / injection_period

    # A single random injection time within the requested time range.
    tinjection0 = numpy.random.RandomState(seed=seed).uniform(tmin, tmax)

    # Calculate all other injection times from the singal random one, knowing
    # injections are `injection_period` apart in time between tmin and tmax.
    tinjections = [tinjection0]
    tinjection = tinjection0
    while tinjection > tmin:
        tinjection -= injection_period
        if tinjection >= tmin:
            tinjections.append(tinjection)
    tinjection = tinjection0
    while tinjection < tmax:
        tinjection += injection_period
        if tinjection <= tmax:
            tinjections.append(tinjection)

    # Add one injection before tmin and one injection after tmax.
    tinjections.append(max(tinjections) + injection_period)
    tinjections.append(min(tinjections) - injection_period)
    tinjections = sorted(tinjections)

    # Calculate the monitor intensity at all points in time
    # based on where each point is between two injections,
    # knowing the intensity decay after each injection.
    monitor = numpy.zeros(npoints)
    for t0, t1 in zip(tinjections[:-1], tinjections[1:]):
        idx = (time >= t0) & (time < t1)
        tperiod = time[idx] - t0
        monitor[idx] = numpy.exp(-decay * tperiod)
    return monitor



import importlib.metadata
from contextlib import ExitStack
from datetime import datetime
from typing import Dict
from typing import Optional

from ewoksdata.data.hdf5.dataset_writer import IndexedDatasetWriter
from ewoksdata.data.nexus import create_nexus_group
from ewoksdata.data.nexus import select_default_plot
from silx.io.dictdump import dicttonx

from ..fit.types import UNITS
from ..fit.types import XRFBatchFitResult
from ..pymca_config import PyMcaXrfConfiguration
from ..pymca_config import pymca_configdict_from_model


class XrfResultWriter:
    _PROGRAM_VERSION = importlib.metadata.version("ewoksfluo")

    def __init__(self, output_parent_uri: str, default_group: Optional[str] = None):
        self._output_parent_uri = output_parent_uri
        self._default_group = default_group

        # Output
        self._already_existed = False
        self._xrf_results_uri = None
        self._output_root_uri = None
        self._dataset_writers = {}

        # Context parameters
        self._reset_context_parameters()

    def _reset_context_parameters(self):
        self._ctx_stack = None
        self._nxprocess = None
        self._results = None
        self._hdf5_groups: Dict[str, dict] = dict()

    @property
    def xrf_results_uri(self) -> Optional[str]:
        return self._xrf_results_uri

    @property
    def output_root_uri(self) -> Optional[str]:
        return self._output_root_uri

    @property
    def already_existed(self) -> bool:
        return self._already_existed

    def _create_group(self, name: str, data: dict) -> None:
        is_nxdata = data["@NX_class"] == "NXdata"
        if is_nxdata:
            if self._results is None:
                self._results = self._nxprocess.create_group("results")
                self._results.attrs["NX_class"] = "NXcollection"
            parent = self._results
        else:
            parent = self._nxprocess
        if name in parent:
            return

        dicttonx(data, parent.file, h5path=f"{parent.name}/{name}")

        if is_nxdata:
            group = parent[name]
            self._hdf5_groups[name] = {"group": group, "signals": list()}
            if name == self._default_group:
                select_default_plot(group)

    def _create_nxdata_group(self, name: str, data: dict) -> None:
        if self._results is None:
            self._results = self._nxprocess.create_group("results")
            self._results.attrs["NX_class"] = "NXcollection"
        parent = self._results
        if name in parent:
            return

        dicttonx(data, parent.file, h5path=f"{parent.name}/{name}")

        group = parent[name]
        self._hdf5_groups[name] = {"group": group, "signals": list()}
        if name == self._default_group:
            select_default_plot(group)

    def save_configuration(self, configuration: PyMcaXrfConfiguration) -> None:
        if "configuration" not in self._hdf5_groups:
            data = {
                "@NX_class": "NXnote",
                "date": datetime.now().astimezone().isoformat(),
                "type": "text/plain",
                "data": pymca_configdict_from_model(configuration).tostring(),
            }
            self._create_group("configuration", data)

    def save_fit_result(self, result: XRFBatchFitResult, start: int, stop: int) -> None:
        for group in ["parameters", "uncertainties", "mass_fractions"]:
            if not result.has_data(group):
                continue
            if group not in self._hdf5_groups:
                data = {"@NX_class": "NXdata"}
                self._create_group(group, data)

            if group == "mass_fractions":
                parameter_names = result.mass_fraction_names
            else:
                parameter_names = result.parameter_names

            for name, values in zip(parameter_names, getattr(result, group)):
                writer = self._get_dataset_writer(group, name)
                writer.add_points(values, list(range(start, stop)))

        if result.has_data("mass_fractions_per_layer"):
            for i, mass_fractions in enumerate(
                result.mass_fractions_per_layer, start=1
            ):
                group = f"mass_fractions_layer{i}"

                if group not in self._hdf5_groups:
                    data = {"@NX_class": "NXdata"}
                    self._create_group(group, data)

                for name, values in zip(result.mass_fraction_names, mass_fractions):
                    writer = self._get_dataset_writer(group, name)
                    writer.add_points(values, list(range(start, stop)))

        if result.has_diagnostics:
            group = "fit"

            if group not in self._hdf5_groups:
                data = {
                    "@NX_class": "NXdata",
                    "@axes": [".", "energy"],
                    "energy": result.raw_energies,
                }
                _add_units_attribute(data, "raw_energies", "energy")
                self._create_group(group, data)
                attrs = {"interpretation": "spectrum"}
            else:
                attrs = None

            writer = self._get_dataset_writer(group, "model", attrs=attrs)
            writer.add_points(result.model, list(range(start, stop)))

            writer = self._get_dataset_writer(group, "data", attrs=attrs)
            writer.add_points(result.spectra, list(range(start, stop)))

            writer = self._get_dataset_writer(group, "residuals", attrs=attrs)
            writer.add_points(result.residuals, list(range(start, stop)))

        if result.has_data("derivatives"):
            if "derivatives" not in self._hdf5_groups:
                data = {
                    "@NX_class": "NXdata",
                    "@axes": ["energy"],
                    "energy": result.raw_energies,
                }
                _add_units_attribute(data, "raw_energies", "energy")
                for name, values in zip(result.parameter_names, result.derivatives):
                    data[name] = values
                    data[f"{name}@interpretation"] = "spectrum"
                self._create_group("derivatives", data)

        if result.has_data("quantification_parameters"):
            if "quantification" not in self._hdf5_groups:
                parameters = result.quantification_parameters._asdict()
                data = {"@NX_class": "NXparameters", **parameters}
                for name in parameters:
                    _add_units_attribute(data, name, name)
                self._create_group("quantification", data)

    def __enter__(self) -> "XrfResultWriter":
        self._ctx_stack = ExitStack().__enter__()
        ctx = create_nexus_group(
            self._output_parent_uri, default_levels=("results", "fit"), retry_timeout=0
        )
        self._nxprocess, self._already_existed = self._ctx_stack.enter_context(ctx)
        self._nxprocess.attrs["NX_class"] = "NXprocess"
        if not self._already_existed:
            self._nxprocess["program"] = "ewoksfluo"
            self._nxprocess["version"] = self._PROGRAM_VERSION
        self._xrf_results_uri = (
            f"{self._nxprocess.file.filename}::{self._nxprocess.name}/results"
        )
        entry_name = [s for s in self._nxprocess.parent.name.split("/") if s][0]
        self._output_root_uri = f"{self._nxprocess.file.filename}::/{entry_name}"
        return self

    def __exit__(self, *args) -> None:
        try:
            self._finalize()
        finally:
            try:
                result = self._ctx_stack.__exit__(*args)
            finally:
                self._reset_context_parameters()
        return result

    def _get_dataset_writer(
        self,
        group: str,
        name: str,
        npoints: Optional[int] = None,
        attrs: Optional[dict] = None,
    ) -> IndexedDatasetWriter:
        key = (group, name)
        if key in self._dataset_writers:
            return self._dataset_writers[key]

        name = name.replace(" ", "_")
        nxdata = self._hdf5_groups[group]

        ctx = IndexedDatasetWriter(
            parent=nxdata["group"], name=name, npoints=npoints, attrs=attrs
        )
        nxdata["signals"].append(name)
        writer = self._ctx_stack.enter_context(ctx)
        self._dataset_writers[key] = writer
        return writer

    def _finalize(self):
        for writer in self._dataset_writers.values():
            writer.flush_buffer()
        for nxdata in self._hdf5_groups.values():
            group = nxdata["group"]
            if "signal" in group.attrs:
                continue
            signals = nxdata["signals"]
            if signals:
                group.attrs["signal"] = signals[0]
            if len(signals) > 1:
                group.attrs["auxiliary_signals"] = signals[1:]


def _add_units_attribute(data: dict, model_name: str, hdf5_name: str):
    if UNITS.get(model_name):
        data[f"{hdf5_name}@units"] = UNITS[model_name]



import logging
import time
from concurrent.futures import Future
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import as_completed
from contextlib import ExitStack
from contextlib import contextmanager
from multiprocessing import Manager
from queue import Queue
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Union

import numpy
from ewoksdata.data.url import as_dataurl
from silx.io import h5py_utils

from ... import resource_utils
from ..fit import fit_xrf_spectra
from ..fit.types import XRFBatchFitResult
from ..pymca_config import PyMcaXrfConfiguration
from ..pymca_config import modify_pymca_configuration
from ..pymca_config import pymca_configdict_from_file
from ..pymca_config import pymca_configdict_to_model
from ..pymca_config import set_beam_energy
from ._progress import ProgressLogger
from ._save import XrfResultWriter
from ._types import FitArguments
from ._types import FitTask
from ._types import FitUris
from ._types import ProgressStats

logger = logging.getLogger(__name__)


def fit_xrf_urls(
    fit_io_uris: List[FitUris],
    configuration: Union[PyMcaXrfConfiguration, str],
    individual_weights: Optional[bool] = None,
    positive_peak_areas: Optional[bool] = None,
    mlines: Optional[dict] = None,
    quantification: Union[Dict[str, float], bool, None] = None,
    energy_multiplier: Optional[float] = None,
    fast_fitting: Optional[bool] = None,
    native_fitting: Optional[bool] = None,
    native_legacy_fitting: Optional[bool] = None,
    diagnostics: Optional[bool] = None,
    max_workers: Optional[int] = None,
    block_size: Optional[int] = None,
) -> None:
    """
    :param fit_io_uris: List of URI's to XRF spectra with shape `(num_spectra, num_channels)`
                        to fit and the associated URI to save the fit results.
    :param configuration: PyMca configuration.
    :param individual_weights: When fitting with weights, use the weight of each
                                spectrum (slow) instead of the average weight.
    :param positive_peak_areas:
    :param mlines: elements (keys) which M line group must be replaced by some M subgroups (values). Defaults to None.
    :param quantification: Calculate mass fractions from peak area's.
    :param energy_multiplier: adds a higher energy bound equal to energy*energy_multiplier to include high-energy peaks. Default: no bound is added.
    :param fast_fitting: Fast fitting means fit all spectra by solving a single linear system of equations.
    :param native_fitting: Use native PyMca batch processing.
    :param native_legacy_fitting: Use legacy native PyMca batch processing.
    :param diagnostics: fit model and residuals.
    :param max_workers: Number of parallel fitting.
    :param block_size: Number of spectra to read and fit at once.
    """

    block_size_max = 1000 if fast_fitting else 100

    if max_workers is None:
        max_workers = resource_utils.get_available_cpus(exclude_current=True)

    if block_size is None:
        # diagnostics = True, fast_fitting = True, native_fitting = False: x9
        # diagnostics = True, fast_fitting = True, native_fitting = True: x4.5
        # diagnostics = True, fast_fitting = False, native_fitting = False: x8.5
        # diagnostics = True, fast_fitting = False, native_fitting = True: 6.5
        shape = (_get_mca_num_channels(fit_io_uris[0].xrf_spectra_uris[0]),)
        block_size = min(
            resource_utils.array_block_size(
                shape, numpy.float64, max_workers, multiplier=9
            ),
            block_size_max,
        )

    logger.info(
        "Fit XRF spectra in blocks of %d in %d workers", block_size, max_workers
    )

    configuration = _initialize_configuration(
        configuration,
        mlines=mlines,
        quantification=quantification,
        fast_fitting=fast_fitting,
    )

    fit_arguments = FitArguments(
        configuration=configuration,
        quantification=bool(quantification),
        individual_weights=individual_weights,
        positive_peak_areas=positive_peak_areas,
        energy_multiplier=energy_multiplier,
        fast_fitting=fast_fitting,
        native_fitting=native_fitting,
        native_legacy_fitting=native_legacy_fitting,
        diagnostics=diagnostics,
    )

    with ExitStack() as stack:
        progress = stack.enter_context(ProgressLogger())

        default_group = "fit" if diagnostics else "parameters"
        writers = {
            fit_io_uri.output_parent_uri: stack.enter_context(
                XrfResultWriter(
                    fit_io_uri.output_parent_uri, default_group=default_group
                )
            )
            for fit_io_uri in fit_io_uris
        }
        for writer in writers.values():
            writer.save_configuration(configuration)

        # Data structures and flow for reading, fitting and writing data:
        # - Generate `FitTask` instances from `FitUris` instances.
        # - Generate `XRFBatchFitResult` instances from `FitTask` instances.
        # - Save `XRFBatchFitResult` instances with to corresponding `XrfResultWriter` instance.
        if max_workers == 0:
            _sequential_execution(
                fit_io_uris,
                fit_arguments,
                writers,
                progress,
                block_size,
            )
        else:
            _parallel_execution(
                fit_io_uris,
                fit_arguments,
                writers,
                stack,
                progress,
                block_size,
                max_workers,
            )


def _sequential_execution(
    fit_io_uris: List[FitUris],
    fit_arguments: FitArguments,
    writers: Dict[str, XrfResultWriter],
    progress: ProgressLogger,
    block_size: int,
):
    for fit_io_uri in fit_io_uris:
        for fit_task in _iter_fit_tasks(fit_io_uri, block_size):
            progress.accumulate(expected_num_spectra=fit_task.num_spectra)
            fit_task, fit_result = _execute_fit_task(fit_task, fit_arguments)
            _finish_task(fit_task, fit_result, writers, progress)


def _parallel_execution(
    fit_io_uris: List[FitUris],
    fit_arguments: FitArguments,
    writers: Dict[str, XrfResultWriter],
    stack: ExitStack,
    progress: ProgressLogger,
    block_size: int,
    max_workers: int,
):
    producers = set()
    producer_pool = stack.enter_context(_pool(max_workers, producers))

    manager = stack.enter_context(Manager())
    fit_tasks = manager.Queue()
    for fit_io_uri in fit_io_uris:
        future = producer_pool.submit(
            _enqueue_fit_tasks, fit_tasks, fit_io_uri, block_size
        )
        producers.add(future)

    fit_futures = set()
    fit_pool = stack.enter_context(_pool(max_workers, fit_futures))

    while producers or not fit_tasks.empty():
        # Submit new fit tasks
        while not fit_tasks.empty():
            fit_task = fit_tasks.get()
            progress.accumulate(expected_num_spectra=fit_task.num_spectra)
            fit_futures.add(fit_pool.submit(_execute_fit_task, fit_task, fit_arguments))

        # Save finished fit results
        for fit_task, fit_result in _purge_futures(fit_futures, "Fitting"):
            _finish_task(fit_task, fit_result, writers, progress)

        # Clear finished producers
        for _ in _purge_futures(producers, "Reading data"):
            pass

    for fit_task, fit_result in _wait_futures(fit_futures, "Fitting"):
        _finish_task(fit_task, fit_result, writers, progress)


def _finish_task(
    fit_task: FitTask,
    fit_result: XRFBatchFitResult,
    writers: Dict[str, XrfResultWriter],
    progress: ProgressLogger,
) -> None:
    t0 = time.perf_counter()
    writer = writers[fit_task.output_parent_uri]
    writer.save_fit_result(fit_result, fit_task.start_out, fit_task.stop_out)
    t1 = time.perf_counter()
    fit_task.progress.seconds_write += t1 - t0
    progress.accumulate(
        read=fit_task.progress.seconds_read,
        process=fit_task.progress.seconds_process,
        write=fit_task.progress.seconds_write,
        processed_num_spectra=fit_result.num_spectra,
    )


@contextmanager
def _pool(
    max_workers: int, futures: Sequence[Future]
) -> Generator[ProcessPoolExecutor, None, None]:
    with ProcessPoolExecutor(max_workers=max_workers) as producer_pool:
        try:
            yield producer_pool
        except Exception:
            for f in futures:
                f.cancel()
            raise


def _initialize_configuration(
    configuration: Union[PyMcaXrfConfiguration, str],
    mlines: Optional[dict] = None,
    quantification: Union[Dict[str, float], bool, None] = None,
    fast_fitting: bool = False,
) -> PyMcaXrfConfiguration:
    if isinstance(configuration, str):
        configuration = pymca_configdict_to_model(
            pymca_configdict_from_file(configuration)
        )
    else:
        configuration = configuration.model_copy(deep=True)

    modify_pymca_configuration(
        configuration,
        energy=None,
        energy_multiplier=None,
        mlines=mlines,
        quantification=quantification,
        fast_fitting=fast_fitting,
    )
    return configuration


def _enqueue_fit_tasks(fit_tasks: Queue, fit_io_uri: FitUris, block_size: int) -> None:
    for fit_task in _iter_fit_tasks(fit_io_uri, block_size):
        fit_tasks.put(fit_task)


def _iter_fit_tasks(
    fit_io_uri: FitUris, block_size: int
) -> Generator[FitTask, None, None]:
    offset = 0
    t0 = time.perf_counter()
    for xrf_spectra_uri, beam_energy_uri in zip(
        fit_io_uri.xrf_spectra_uris, fit_io_uri.beam_energy_uris
    ):
        stop_in = None
        for start_in, stop_in in _iter_scan_slices(xrf_spectra_uri, block_size):
            t1 = time.perf_counter()
            fit_task = FitTask(
                xrf_spectra_uri=xrf_spectra_uri,
                beam_energy_uri=beam_energy_uri,
                start_in=start_in,
                stop_in=stop_in,
                start_out=start_in + offset,
                stop_out=stop_in + offset,
                output_parent_uri=fit_io_uri.output_parent_uri,
                progress=ProgressStats(),
            )
            fit_task.progress.num_spectra = fit_task.num_spectra
            fit_task.progress.seconds_read += t1 - t0
            yield fit_task
        if stop_in is not None:
            offset = stop_in


def _purge_futures(futures: Set[Future], task: str) -> Generator[Any, None, None]:
    done = [f for f in futures if f.done()]
    for f in done:
        try:
            yield f.result()
        except Exception as ex:
            logger.error("%s failed: %s", task, ex)
            raise
        futures.remove(f)


def _wait_futures(futures: Set[Future], task: str) -> Generator[Any, None, None]:
    for f in as_completed(futures):
        try:
            yield f.result()
        except Exception as ex:
            logger.error("%s failed: %s", task, ex)
            raise
        futures.remove(f)


def _execute_fit_task(
    fit_task: FitTask, fit_arguments: FitArguments
) -> Tuple[FitTask, XRFBatchFitResult]:

    t0 = time.perf_counter()
    spectra, energy = _read_data(
        fit_task.xrf_spectra_uri,
        fit_task.beam_energy_uri,
        fit_task.start_in,
        fit_task.stop_in,
    )
    t1 = time.perf_counter()

    if energy is None or energy.ndim == 0:
        pass
    elif energy.ndim == 1:
        energy = energy[0]
    else:
        indices = numpy.unique(energy)
        if len(indices) != 1:
            raise NotImplementedError(
                f"Block has more than on primary beam energy: {energy[indices]}"
            )
        energy = energy[indices[0]]

    fit_task.progress.seconds_read += t1 - t0

    if energy is not None:
        configuration = _modify_configuration(
            fit_arguments.configuration, energy, fit_arguments.energy_multiplier
        )

    fit_result = fit_xrf_spectra(
        spectra,
        configuration=configuration,
        quantification=fit_arguments.quantification,
        individual_weights=fit_arguments.individual_weights,
        positive_peak_areas=fit_arguments.positive_peak_areas,
        diagnostics=fit_arguments.diagnostics,
        fast_fitting=fit_arguments.fast_fitting,
        native_fitting=fit_arguments.native_fitting,
        native_legacy_fitting=fit_arguments.native_legacy_fitting,
    )

    t2 = time.perf_counter()

    fit_task.progress.seconds_process += t2 - t1
    return fit_task, fit_result


def _modify_configuration(
    configuration: PyMcaXrfConfiguration,
    energy: float,
    energy_multiplier: Optional[float],
) -> PyMcaXrfConfiguration:
    configuration = configuration.model_copy(deep=True)
    set_beam_energy(configuration, energy, energy_multiplier)
    return configuration


def _iter_scan_slices(
    spectra_uri: str, block_size: int
) -> Generator[Tuple[int, int], None, None]:
    # TODO: assumes static HDF5 file for now
    url = as_dataurl(spectra_uri)
    with h5py_utils.File(url.file_path()) as h5f:
        nspectra = len(h5f[url.data_path()])
    for start in range(0, nspectra, block_size):
        yield start, min(start + block_size, nspectra)


def _get_mca_num_channels(xrf_spectra_uri: str) -> int:
    spectra_url = as_dataurl(xrf_spectra_uri)
    with h5py_utils.File(spectra_url.file_path()) as h5f_spectra:
        dset_spectra = h5f_spectra[spectra_url.data_path()]
        return dset_spectra.shape[-1]


def _read_data(
    xrf_spectra_uri: str, beam_energy_uri: Optional[str], start: int, stop: int
) -> Generator[Tuple[numpy.ndarray, Optional[numpy.ndarray]], None, None]:
    """Yields arrays with shape `(Nspectra, Nchan)`."""
    spectra_url = as_dataurl(xrf_spectra_uri)
    if beam_energy_uri:
        energy_url = as_dataurl(beam_energy_uri)

        with h5py_utils.File(spectra_url.file_path()) as h5f_spectra:
            with h5py_utils.File(energy_url.file_path()) as h5f_energy:
                dset_spectra = h5f_spectra[spectra_url.data_path()]
                spectra = dset_spectra[start:stop]
                dset_energy = h5f_energy[energy_url.data_path()]
                if dset_energy.size in (0, 1):
                    energy = dset_energy[()]
                else:
                    energy = dset_energy[start:stop]
                return spectra, energy
    else:
        with h5py_utils.File(spectra_url.file_path()) as h5f_spectra:
            dset_spectra = h5f_spectra[spectra_url.data_path()]
            spectra = dset_spectra[start:stop]
            return spectra, None



import logging
import os
from contextlib import contextmanager
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Dict
from typing import Generator
from typing import Optional
from typing import Union

from PyMca5.PyMcaIO.ConfigDict import ConfigDict

from ._modify.beam_energy import set_beam_energy
from ._modify.fast_fitting import force_fast_fitting
from ._modify.fit_weights import set_fit_weights
from ._modify.quantification import enable_quantification
from ._modify.split_mlines import split_mlines
from ._modify.summary import configuration_summary
from .model import PyMcaXrfConfiguration

logger = logging.getLogger(__name__)


def pymca_configdict_from_model(configuration: PyMcaXrfConfiguration) -> ConfigDict:
    data = configuration.model_dump(mode="python", exclude_unset=True)
    return ConfigDict(initdict=data)


def pymca_configdict_to_model(
    configuration: Union[ConfigDict, dict],
) -> PyMcaXrfConfiguration:
    return PyMcaXrfConfiguration(**configuration)


def pymca_configdict_from_file(filename: Union[str, Path]) -> ConfigDict:
    return ConfigDict(filelist=[str(filename)])


def pymca_configdict_to_file(
    configuration: ConfigDict, filename: Union[str, Path]
) -> None:
    configuration.write(str(filename))


@contextmanager
def temporary_pymca_configdict_to_file(
    configuration: ConfigDict,
) -> Generator[str, None, None]:
    tmp = NamedTemporaryFile(suffix=".cfg", delete=False)
    try:
        tmp.close()
        pymca_configdict_to_file(configuration, tmp.name)
        yield tmp.name
    finally:
        try:
            os.remove(tmp.name)
        except FileNotFoundError:
            pass


def modify_pymca_configuration(
    configuration: PyMcaXrfConfiguration,
    energy: Optional[float] = None,
    energy_multiplier: Optional[float] = None,
    mlines: Optional[dict] = None,
    quantification: Optional[Dict[str, float]] = None,
    fast_fitting: bool = False,
) -> None:
    """
    :param configuration: PyMca configuration to modify in-place.
    :param energy: Primary beam energy in keV/
    :param energy_multiplier: Add high primary energy with given multiplier with very low weight.
    :param mlines: Elements (keys) which M line group must be replaced by some M subgroups (values).
    :param quantification: Enable calculation of elemental concentrations when `True` or a dictionary. `None` defaults to `True`.
    :param fast_fitting: Fast fitting means fit all spectra by solving a single linear system of equations.
    """
    set_beam_energy(configuration, energy, energy_multiplier)

    if quantification is False:
        quantification = None
    elif quantification is None or quantification is True:
        quantification = {}
    if quantification is not None:
        enable_quantification(configuration, quantification)

    set_fit_weights(configuration, quantification)

    if fast_fitting:
        force_fast_fitting(configuration)

    if mlines is not None:
        split_mlines(configuration, mlines)

    summary = configuration_summary(configuration, quantification, fast_fitting)
    logger.info("XRF fit configuration summary:\n %s", "\n ".join(summary))



import sys
from typing import Dict
from typing import Sequence
from typing import Union

if sys.version_info < (3, 12):
    import importlib_resources as resources
else:
    from importlib import resources

import numpy

from ..xrffit import pymca_config


def test_round_trip():
    files = resources.files("PyMca5.PyMcaData")
    with resources.as_file(files) as pkg_path:
        for cfg_file in pkg_path.glob("*.cfg"):
            config_dict_org = pymca_config.pymca_configdict_from_file(cfg_file)
            model = pymca_config.pymca_configdict_to_model(config_dict_org)
            config_dict = pymca_config.pymca_configdict_from_model(model)
            assert config_dict == config_dict_org, cfg_file


def _generate_pymca_config(fit_config: Dict[str, Sequence[Union[float, int]]]):
    files = resources.files("PyMca5.PyMcaData")
    with resources.as_file(files) as pkg_path:
        config_dict = pymca_config.pymca_configdict_from_file(pkg_path / "Steel.cfg")
        config_dict["fit"].update(fit_config)
        return pymca_config.pymca_configdict_to_model(config_dict)


def test_replace_energy():
    configuration = _generate_pymca_config(
        {
            "energy": [2, 5, 10, 20, 50],
            "energyflag": [1, 0, 1, 0, 0],
            "energyscatter": [1, 0, 0, 1, 0],
            "energyweight": [1.0, 0.25, 0.2, 0.1, 0.1],
        }
    )
    energy = 2.5
    energy_multiplier = 10

    pymca_config.set_beam_energy(configuration, energy, energy_multiplier)

    numpy.testing.assert_allclose(
        configuration.fit.energy, numpy.array([2.5, 12.5, 25])
    )
    numpy.testing.assert_equal(configuration.fit.energyflag, numpy.array([1, 1, 1]))
    numpy.testing.assert_equal(configuration.fit.energyscatter, numpy.array([1, 0, 0]))
    numpy.testing.assert_allclose(
        configuration.fit.energyweight, numpy.array([1.0, 0.2, 1e-10])
    )


def test_replace_energy_all_defined():
    configuration = _generate_pymca_config(
        {
            "energy": [2],
            "energyflag": [1],
            "energyscatter": [1],
            "energyweight": [1.0],
        }
    )
    energy = 4
    energy_multiplier = 10

    pymca_config.set_beam_energy(configuration, energy, energy_multiplier)

    numpy.testing.assert_allclose(configuration.fit.energy, numpy.array([4, 40]))
    numpy.testing.assert_equal(configuration.fit.energyflag, numpy.array([1, 1]))
    numpy.testing.assert_equal(configuration.fit.energyscatter, numpy.array([1, 0]))
    numpy.testing.assert_allclose(
        configuration.fit.energyweight, numpy.array([1.0, 1e-10])
    )


def test_replace_energy_no_multiplier():
    configuration = _generate_pymca_config(
        {
            "energy": [2, 5, 10, 20, 50],
            "energyflag": [1, 0, 1, 0, 0],
            "energyscatter": [1, 0, 0, 1, 0],
            "energyweight": [1.0, 0.25, 0.2, 0.1, 0.1],
        }
    )
    energy = 2.5
    energy_multiplier = None

    pymca_config.set_beam_energy(configuration, energy, energy_multiplier)

    numpy.testing.assert_allclose(configuration.fit.energy, numpy.array([2.5, 12.5]))
    numpy.testing.assert_equal(configuration.fit.energyflag, numpy.array([1, 1]))
    numpy.testing.assert_equal(configuration.fit.energyscatter, numpy.array([1, 0]))
    numpy.testing.assert_allclose(
        configuration.fit.energyweight, numpy.array([1.0, 0.2])
    )


def test_replace_energy_not_defined():
    configuration = _generate_pymca_config(
        {
            "energy": [2, 5, 10, 20, 50],
            "energyflag": [0, 0, 0, 0, 0],
            "energyscatter": [1, 0, 0, 1, 0],
            "energyweight": [1.0, 0.25, 0.2, 0.1, 0.1],
        }
    )
    energy = 6
    energy_multiplier = 3

    pymca_config.set_beam_energy(configuration, energy, energy_multiplier)

    numpy.testing.assert_allclose(
        configuration.fit.energy, numpy.array([energy, energy_multiplier * energy])
    )
    numpy.testing.assert_equal(configuration.fit.energyflag, numpy.array([1, 1]))
    numpy.testing.assert_equal(configuration.fit.energyscatter, numpy.array([0, 0]))
    numpy.testing.assert_allclose(
        configuration.fit.energyweight, numpy.array([1.0, 1e-10])
    )


def test_replace_energy_not_defined_no_multiplier():
    configuration = _generate_pymca_config(
        {
            "energy": [2, 5, 10, 20, 50],
            "energyflag": [0, 0, 0, 0, 0],
            "energyscatter": [1, 0, 0, 1, 0],
            "energyweight": [1.0, 0.25, 0.2, 0.1, 0.1],
        }
    )
    energy = 7.1
    energy_multiplier = None

    pymca_config.set_beam_energy(configuration, energy, energy_multiplier)

    numpy.testing.assert_allclose(configuration.fit.energy, numpy.array([energy]))
    numpy.testing.assert_equal(configuration.fit.energyflag, numpy.array([1]))
    numpy.testing.assert_equal(configuration.fit.energyscatter, numpy.array([0]))
    numpy.testing.assert_allclose(configuration.fit.energyweight, numpy.array([1.0]))


def test_replace_multiplier_only():
    configuration = _generate_pymca_config(
        {
            "energy": [4],
            "energyflag": [1],
            "energyscatter": [1],
            "energyweight": [1.0],
        }
    )
    energy_multiplier = 10

    pymca_config.set_beam_energy(configuration, None, energy_multiplier)

    numpy.testing.assert_allclose(configuration.fit.energy, numpy.array([4, 40]))
    numpy.testing.assert_equal(configuration.fit.energyflag, numpy.array([1, 1]))
    numpy.testing.assert_equal(configuration.fit.energyscatter, numpy.array([1, 0]))
    numpy.testing.assert_allclose(
        configuration.fit.energyweight, numpy.array([1.0, 1e-10])
    )


def test_replace_multiplier_only_no_energy():
    configuration = _generate_pymca_config(
        {
            "energy": [],
            "energyflag": [],
            "energyscatter": [],
            "energyweight": [],
        }
    )
    energy = None
    energy_multiplier = 10

    pymca_config.set_beam_energy(configuration, energy, energy_multiplier)

    numpy.testing.assert_allclose(configuration.fit.energy, numpy.array([]))
    numpy.testing.assert_equal(configuration.fit.energyflag, numpy.array([]))
    numpy.testing.assert_equal(configuration.fit.energyscatter, numpy.array([]))
    numpy.testing.assert_allclose(configuration.fit.energyweight, numpy.array([]))



from contextlib import contextmanager

import fabio
import h5py
import numpy
import pytest
from ewoksorange.tests.utils import execute_task

from orangecontrib.ewoksfluo.categories.input.spec_to_bliss import OWSpecToBliss

from ..io.convert import spec_to_bliss
from ..tasks.example_data.deadtime import apply_extendable_deadtime


def test_spec_to_bliss(tmp_path):
    with _assert_spec_to_bliss(tmp_path) as (spec_filename, bliss_filename, scans):
        spec_to_bliss(spec_filename, bliss_filename, scans, mode="w")


@pytest.mark.parametrize("fluoxas", [True, False])
def test_xia_spec_to_bliss(tmp_path, fluoxas):
    with _assert_spec_to_bliss_with_xia(tmp_path, fluoxas=fluoxas) as (
        spec_filename,
        bliss_filename,
        scans,
    ):
        spec_to_bliss(spec_filename, bliss_filename, scans, mode="w")


def test_spec_to_bliss_task(tmp_path):
    with _assert_spec_to_bliss(tmp_path) as (spec_filename, bliss_filename, scans):
        inputs = {
            "input_filename": spec_filename,
            "output_filename": bliss_filename,
            "scans": scans,
        }
        execute_task(OWSpecToBliss.ewokstaskclass, inputs)


def test_spec_to_bliss_widget(tmp_path, qtapp):
    with _assert_spec_to_bliss(tmp_path) as (spec_filename, bliss_filename, scans):
        inputs = {
            "input_filename": spec_filename,
            "output_filename": bliss_filename,
            "scans": scans,
        }
        execute_task(OWSpecToBliss, inputs, timeout=10)


@pytest.mark.parametrize("fluoxas", [True, False])
def test_xia_spec_to_bliss_task(tmp_path, fluoxas):
    with _assert_spec_to_bliss_with_xia(tmp_path, fluoxas=fluoxas) as (
        spec_filename,
        bliss_filename,
        scans,
    ):
        inputs = {
            "input_filename": spec_filename,
            "output_filename": bliss_filename,
            "scans": scans,
        }
        execute_task(OWSpecToBliss.ewokstaskclass, inputs)


@pytest.mark.parametrize("fluoxas", [True, False])
def test_xia_spec_to_bliss_widget(tmp_path, fluoxas, qtapp):
    with _assert_spec_to_bliss_with_xia(tmp_path, fluoxas=fluoxas) as (
        spec_filename,
        bliss_filename,
        scans,
    ):
        inputs = {
            "input_filename": spec_filename,
            "output_filename": bliss_filename,
            "scans": scans,
        }
        execute_task(OWSpecToBliss, inputs)


@contextmanager
def _assert_spec_to_bliss(tmp_path):
    spec_filename = str(tmp_path / "spec.dat")
    bliss_filename = str(tmp_path / "spec.h5")
    samx = numpy.random.randint(0, 100, size=11)
    I0 = numpy.random.randint(0, 100, size=11)
    IC = numpy.random.randint(0, 100, size=11)
    samz = numpy.arange(11)

    with open(spec_filename, mode="w") as f:
        f.write(f"#F {spec_filename}\n")
        f.write("\n")
        f.write("#O0 samx  samy\n")
        f.write("\n")
        f.write("#S 1 ascan samz 0 10 10 0.1\n")
        f.write("#P0 1.2  2.3\n")
        f.write("#N 100\n")
        f.write("#L samx  zap_p201_I0  zap_p201_IC\n")
        for line in numpy.stack([samx, I0, IC], axis=1):
            f.write("  ".join(list(map(str, line))) + "\n")

    yield spec_filename, bliss_filename, 1

    with h5py.File(bliss_filename, mode="r") as f:
        data = f["/1.1/measurement/samx"][()]
        numpy.testing.assert_array_equal(data, samx)
        data = f["/1.1/instrument/samx/value"][()]
        numpy.testing.assert_array_equal(data, samx)
        data = f["/1.1/instrument/positioners/samx"][()]
        numpy.testing.assert_array_equal(data, samx)

        data = f["/1.1/measurement/samz"][()]
        numpy.testing.assert_array_equal(data, samz)
        data = f["/1.1/instrument/samz/value"][()]
        numpy.testing.assert_array_equal(data, samz)
        data = f["/1.1/instrument/positioners/samz"][()]
        numpy.testing.assert_array_equal(data, samz)

        data = f["/1.1/measurement/I0"][()]
        numpy.testing.assert_array_equal(data, I0)
        data = f["/1.1/instrument/I0/data"][()]
        numpy.testing.assert_array_equal(data, I0)

        data = f["/1.1/measurement/IC"][()]
        numpy.testing.assert_array_equal(data, IC)
        data = f["/1.1/instrument/IC/data"][()]
        numpy.testing.assert_array_equal(data, IC)

        assert f["/1.1/instrument/positioners_start/samx"][()] == 1.2
        assert f["/1.1/instrument/positioners_start/samy"][()] == 2.3
        assert f["/1.1/instrument/positioners/samy"][()] == 2.3


@contextmanager
def _assert_spec_to_bliss_with_xia(tmp_path, fluoxas: bool = False):
    spec_filename = str(tmp_path / "spec.dat")
    bliss_filename = str(tmp_path / "spec.h5")

    with open(spec_filename, mode="w") as f:
        f.write(f"#F {spec_filename}\n")
        f.write("\n")
        f.write("#O0 samx  samy\n")
        f.write("\n")
        if fluoxas:
            f.write("#S 1 fXAS_scan 17.0 17.0\n")
        else:
            f.write("#S 1 zapimage  sampy 0 5 6 100 sampz 0 4 4 0\n")

        f.write("#P0 1.2  2.3\n")
        if fluoxas:
            f.write("#N 2\n")
            f.write("#L specctr1  specctr2\n")
            f.write("0 0")  # only to know how many maps in fXAS_scan
        else:
            f.write("#N 0\n")
            f.write("#L\n")
        f.write("#C \n")

        if fluoxas:
            radix = "radix_0001"
            xiadir = tmp_path / radix
            f.write(f"#C DIRECTORY        : {str(xiadir)}\n")
        else:
            radix = "radix"
            xiadir = tmp_path / "xia"
            f.write(f"#C DIRECTORY        : {str(xiadir)}\n")
            f.write(f"#C RADIX            : {radix}\n")
            f.write("#C ZAP SCAN NUMBER  : 10\n")
            f.write("#C ZAP IMAGE NUMBER : 0\n")
        xiadir.mkdir()

    nrows = 5
    ncols = 6
    ndet = 2
    nchan = 7

    # Simulate a Dual-Channel signal processor with paralyzable dead time
    photons = numpy.random.randint(1000, 5000, size=(nrows, ncols, ndet))
    elapsed_time = numpy.ones((nrows, ncols, ndet))  # seconds
    triggers = apply_extendable_deadtime(photons, elapsed_time, 0.1)
    events = apply_extendable_deadtime(photons, elapsed_time, 20)
    dt_triggers = (photons - triggers) / photons
    dt_events = (photons - events) / photons
    trigger_live_time = elapsed_time - dt_triggers * elapsed_time
    live_time = elapsed_time - dt_events * elapsed_time
    trigger_count_rate = triggers / trigger_live_time
    event_count_rate = events / elapsed_time

    # XIA stores these at unit32
    detnr = numpy.zeros((nrows, ncols, ndet), dtype=numpy.uint32)
    for det in range(ndet):
        detnr[..., det] = det
    xiast = numpy.stack(
        [
            detnr,
            events,
            (trigger_count_rate + 0.5).astype(numpy.uint32),
            (event_count_rate + 0.5).astype(numpy.uint32),
            (trigger_live_time * 1000 + 0.5).astype(numpy.uint32),
            (dt_events * 100 + 0.5).astype(numpy.uint32),
        ],
        axis=-1,
    )
    xiast = xiast.reshape((nrows, ncols, ndet * 6))

    # MCA and counters
    xiamca = numpy.random.randint(1000, 5000, size=(nrows, ndet, ncols, nchan))
    tika = numpy.random.randint(1000, 5000, size=(ndet, nrows, ncols))
    I0 = numpy.random.randint(1000, 5000, size=(nrows, ncols))
    IC = numpy.random.randint(1000, 5000, size=(nrows, ncols))
    if not fluoxas:
        sampy = numpy.tile(numpy.linspace(0.5, ncols - 1.5, ncols), nrows)
        sampz = numpy.repeat(numpy.arange(nrows), ncols)

    # Save as EDF's
    header = {"energy": 17.0}
    for line, (stdata, mcadata) in enumerate(zip(xiast, xiamca)):
        filename = str(xiadir / f"{radix}_xiast_{10:04d}_{0:04d}_{line:04d}.edf")
        _save_edf(filename, stdata, header)
        for det, data in enumerate(mcadata):
            filename = str(
                xiadir / f"{radix}_xia{det:02d}_{10:04d}_{0:04d}_{line:04d}.edf"
            )
            _save_edf(filename, data, header)
    for det, data in enumerate(tika):
        filename = str(xiadir / f"{radix}_xmap_tika{det:02d}_{10:04d}_{0:04d}.edf")
        _save_edf(filename, data)
    filename = str(xiadir / f"{radix}_zap_p201_I0_{10:04d}_{0:04d}.edf")
    _save_edf(filename, I0)
    filename = str(xiadir / f"{radix}_zap_p201_IC_{10:04d}_{0:04d}.edf")
    _save_edf(filename, IC)
    image = numpy.zeros((10, 11))
    for islow in range(nrows):
        for ifast in range(ncols):
            filename = str(
                xiadir / f"{radix}_frelon2_{10:04d}_{islow:04d}_{ifast:04d}.edf"
            )
            image[0, 0] = islow * ncols + ifast
            _save_edf(filename, image)

    # Convert to HDF5
    yield spec_filename, bliss_filename, 1

    # Validate the conversion
    with h5py.File(bliss_filename, mode="r") as f:
        assert f["/1.1/instrument/positioners_start/samx"][()] == 1.2
        assert f["/1.1/instrument/positioners_start/samy"][()] == 2.3
        assert f["/1.1/instrument/positioners_start/energy"][()] == 17.0
        assert f["/1.1/instrument/positioners/samx"][()] == 1.2
        assert f["/1.1/instrument/positioners/samy"][()] == 2.3
        assert f["/1.1/instrument/positioners/energy"][()] == 17.0

        if not fluoxas:
            data = f["/1.1/measurement/sampz"][()]
            numpy.testing.assert_array_equal(data, sampz)
            data = f["/1.1/instrument/sampz/value"][()]
            numpy.testing.assert_array_equal(data, sampz)
            data = f["/1.1/instrument/positioners/sampz"][()]
            numpy.testing.assert_array_equal(data, sampz)

            data = f["/1.1/measurement/sampy"][()]
            numpy.testing.assert_array_equal(data, sampy)
            data = f["/1.1/instrument/sampy/value"][()]
            numpy.testing.assert_array_equal(data, sampy)
            data = f["/1.1/instrument/positioners/sampy"][()]
            numpy.testing.assert_array_equal(data, sampy)

        data = f["/1.1/measurement/I0"][()]
        numpy.testing.assert_array_equal(data, I0.flatten())
        data = f["/1.1/instrument/I0/data"][()]
        numpy.testing.assert_array_equal(data, I0.flatten())

        data = f["/1.1/measurement/IC"][()]
        numpy.testing.assert_array_equal(data, IC.flatten())
        data = f["/1.1/instrument/IC/data"][()]
        numpy.testing.assert_array_equal(data, IC.flatten())

        for det, roi in enumerate(tika):
            data = f[f"/1.1/instrument/xia_det{det}/tika"][()]
            numpy.testing.assert_array_equal(data, roi.flatten())
            data = f[f"/1.1/measurement/xia_det{det}_tika"][()]
            numpy.testing.assert_array_equal(data, roi.flatten())

        for det in range(ndet):
            data = f[f"/1.1/instrument/xia_det{det}/events"][()]
            numpy.testing.assert_array_equal(data, events[..., det].flatten())
            data = f[f"/1.1/measurement/xia_det{det}_events"][()]
            numpy.testing.assert_array_equal(data, events[..., det].flatten())

            data = f[f"/1.1/instrument/xia_det{det}/triggers"][()]
            numpy.testing.assert_allclose(data, triggers[..., det].flatten(), rtol=1e-3)
            data = f[f"/1.1/measurement/xia_det{det}_triggers"][()]
            numpy.testing.assert_allclose(data, triggers[..., det].flatten(), rtol=1e-3)

            data = f[f"/1.1/instrument/xia_det{det}/event_count_rate"][()]
            numpy.testing.assert_allclose(data, event_count_rate[..., det].flatten())
            data = f[f"/1.1/measurement/xia_det{det}_event_count_rate"][()]
            numpy.testing.assert_allclose(data, event_count_rate[..., det].flatten())

            data = f[f"/1.1/instrument/xia_det{det}/trigger_count_rate"][()]
            numpy.testing.assert_allclose(
                data, trigger_count_rate[..., det].flatten(), rtol=1e-3
            )
            data = f[f"/1.1/measurement/xia_det{det}_trigger_count_rate"][()]
            numpy.testing.assert_allclose(
                data, trigger_count_rate[..., det].flatten(), rtol=1e-3
            )

            data = f[f"/1.1/instrument/xia_det{det}/live_time"][()]
            numpy.testing.assert_allclose(data, live_time[..., det].flatten())
            data = f[f"/1.1/measurement/xia_det{det}_live_time"][()]
            numpy.testing.assert_allclose(data, live_time[..., det].flatten())

            data = f[f"/1.1/instrument/xia_det{det}/trigger_live_time"][()]
            numpy.testing.assert_allclose(
                data, trigger_live_time[..., det].flatten(), rtol=1e-3
            )
            data = f[f"/1.1/measurement/xia_det{det}_trigger_live_time"][()]
            numpy.testing.assert_allclose(
                data, trigger_live_time[..., det].flatten(), rtol=1e-3
            )

            data = f[f"/1.1/instrument/xia_det{det}/elapsed_time"][()]
            numpy.testing.assert_allclose(data, elapsed_time[..., det].flatten())
            data = f[f"/1.1/measurement/xia_det{det}_elapsed_time"][()]
            numpy.testing.assert_allclose(data, elapsed_time[..., det].flatten())

            data = f[f"/1.1/instrument/xia_det{det}/fractional_dead_time"][()]
            numpy.testing.assert_allclose(
                data, dt_events[..., det].flatten(), atol=0.01
            )
            data = f[f"/1.1/measurement/xia_det{det}_fractional_dead_time"][()]
            numpy.testing.assert_allclose(
                data, dt_events[..., det].flatten(), atol=0.01
            )

        data = f["/1.1/measurement/frelon2"][()]
        numpy.testing.assert_array_equal(data[:, 0, 0], range(ncols * nrows))


def _save_edf(filename, data, header=None):
    edf = fabio.edfimage.EdfImage(data=data, header=header)
    edf.write(filename)
